import numpy as np

from pythermalcomfort.models import a_pmv


def test_a_pmv():
    np.testing.assert_equal(
        a_pmv([24, 30], 30, vr=0.22, rh=50, met=1.4, clo=0.5, a_coefficient=0.293),
        [0.48, 1.09],
    )



from pythermalcomfort.models import heat_index


def test_heat_index():
    assert heat_index(25, 50) == 25.9
    assert heat_index(77, 50, units="IP") == 78.6
    assert heat_index(30, 80) == 37.7
    assert heat_index(86, 80, units="IP") == 99.8



import numpy as np

from pythermalcomfort.models import clo_tout


def test_clo_tout():
    assert (clo_tout(tout=80.6, units="ip")) == 0.46
    np.testing.assert_equal(clo_tout(tout=[80.6, 82], units="ip"), [0.46, 0.46])
    assert (clo_tout(tout=27)) == 0.46
    np.testing.assert_equal(clo_tout(tout=[27, 24]), [0.46, 0.48])



from pythermalcomfort.models import humidex


def test_humidex():
    assert humidex(25, 50) == {"humidex": 28.2, "discomfort": "Little or no discomfort"}
    assert humidex(30, 80) == {
        "humidex": 43.3,
        "discomfort": "Intense discomfort; avoid exertion",
    }
    assert humidex(31.6, 57.1) == {
        "humidex": 40.8,
        "discomfort": "Intense discomfort; avoid exertion",
    }



from pythermalcomfort.models import solar_gain
import json
import requests

# get file containing validation tables
url = "https://raw.githubusercontent.com/FedericoTartarini/validation-data-comfort-models/main/validation_data.json"
resp = requests.get(url)
reference_tables = json.loads(resp.text)


def test_solar_gain():
    for table in reference_tables["reference_data"]["solar_gain"]:
        for entry in table["data"]:
            inputs = entry["inputs"]
            outputs = entry["outputs"]
            sg = solar_gain(
                inputs["alt"],
                inputs["sharp"],
                inputs["I_dir"],
                inputs["t_sol"],
                inputs["f_svv"],
                inputs["f_bes"],
                inputs["asa"],
                inputs["posture"],
            )
            assert sg["erf"] == outputs["erf"]
            assert sg["delta_mrt"] == outputs["t_rsw"]



import numpy as np

from pythermalcomfort.models import two_nodes
from pythermalcomfort.utilities import (
    units_converter,
    check_standard_compliance_array,
    body_surface_area,
)


def use_fans_heatwaves(
    tdb,
    tr,
    v,
    rh,
    met,
    clo,
    wme=0,
    body_surface_area=1.8258,
    p_atm=101325,
    body_position="standing",
    units="SI",
    max_skin_blood_flow=80,
    **kwargs,
):
    """It helps you to estimate if the conditions you have selected would cause
    heat strain. This occurs when either the following variables reaches its
    maximum value:

    * m_rsw Rate at which regulatory sweat is generated, [mL/h/m2]
    * w : Skin wettedness, adimensional. Ranges from 0 and 1.
    * m_bl : Skin blood flow [kg/h/m2]

    Parameters
    ----------
    tdb : float
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    v : float
        air speed, default in [m/s] in [fps] if `units` = 'IP'
    rh : float
        relative humidity, [%]
    met : float
        metabolic rate, [met]
    clo : float
        clothing insulation, [clo]
    wme : float
        external work, [met] default 0
    body_surface_area : float
        body surface area, default value 1.8258 [m2] in [ft2] if `units` = 'IP'

        The body surface area can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.body_surface_area`.
    p_atm : float
        atmospheric pressure, default value 101325 [Pa] in [atm] if `units` = 'IP'
    body_position: str default="standing"
        select either "sitting" or "standing"
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.
    max_skin_blood_flow : float, [kg/h/m2] default 80
        maximum blood flow from the core to the skin

    Other Parameters
    ----------------
    max_sweating: float, [mL/h/m2] default 500
        max sweating
    round: boolean, default True
        if True rounds output value, if False it does not round it
    limit_inputs : boolean default True
        By default, if the inputs are outside the standard applicability limits the
        function returns nan. If False returns pmv and ppd values even if input values are
        outside the applicability limits of the model.

        The applicability limits are 20 < tdb [°C] < 50, 20 < tr [°C] < 50,
        0.1 < v [m/s] < 4.5, 0.7 < met [met] < 2, and 0 < clo [clo] < 1.

    Returns
    -------
    e_skin : float
        Total rate of evaporative heat loss from skin, [W/m2]. Equal to e_rsw + e_diff
    e_rsw : float
        Rate of evaporative heat loss from sweat evaporation, [W/m2]
    e_diff : float
        Rate of evaporative heat loss from moisture diffused through the skin, [W/m2]
    e_max : float
        Maximum rate of evaporative heat loss from skin, [W/m2]
    q_sensible : float
        Sensible heat loss from skin, [W/m2]
    q_skin : float
        Total rate of heat loss from skin, [W/m2]. Equal to q_sensible + e_skin
    q_res : float
        Total rate of heat loss through respiration, [W/m2]
    t_core : float
        Core temperature, [°C]
    t_skin : float
        Skin temperature, [°C]
    m_bl : float
        Skin blood flow, [kg/h/m2]
    m_rsw : float
        Rate at which regulatory sweat is generated, [mL/h/m2]
    w : float
        Skin wettedness, adimensional. Ranges from 0 and 1.
    w_max : float
        Skin wettedness (w) practical upper limit, adimensional. Ranges from 0 and 1.
    heat_strain : bool
        True if the model predict that the person may be experiencing heat strain
    heat_strain_blood_flow : bool
        True if heat strain is caused by skin blood flow (m_bl) reaching its maximum value
    heat_strain_w : bool
        True if heat strain is caused by skin wettedness (w) reaching its maximum value
    heat_strain_sweating : bool
        True if heat strain is caused by regulatory sweating (m_rsw) reaching its
        maximum value
    """
    # If the SET function is used to calculate the cooling effect then the h_c is
    # calculated in a slightly different way
    default_kwargs = {"round": True, "max_sweating": 500, "limit_inputs": True}
    kwargs = {**default_kwargs, **kwargs}

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    if units.lower() == "ip":
        if body_surface_area == 1.8258:
            body_surface_area = 19.65
        if p_atm == 101325:
            p_atm = 1
        tdb, tr, v, body_surface_area, p_atm = units_converter(
            tdb=tdb, tr=tr, v=v, area=body_surface_area, pressure=p_atm
        )

    output = two_nodes(
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atmospheric=p_atm,
        body_position=body_position,
        max_skin_blood_flow=max_skin_blood_flow,
        round=False,
        output="all",
        max_sweating=kwargs["max_sweating"],
    )

    output_vars = [
        "e_skin",
        "e_rsw",
        "e_max",
        "q_sensible",
        "q_skin",
        "q_res",
        "t_core",
        "t_skin",
        "m_bl",
        "m_rsw",
        "w",
        "w_max",
        "heat_strain_blood_flow",
        "heat_strain_w",
        "heat_strain_sweating",
        "heat_strain",
    ]

    output["heat_strain_blood_flow"] = np.where(
        output["m_bl"] == max_skin_blood_flow, True, False
    )
    output["heat_strain_w"] = np.where(output["w"] == output["w_max"], True, False)
    output["heat_strain_sweating"] = np.where(
        output["m_rsw"] == kwargs["max_sweating"], True, False
    )

    output["heat_strain"] = np.any(
        [
            output["heat_strain_blood_flow"],
            output["heat_strain_w"],
            output["heat_strain_sweating"],
        ],
        axis=0,
    )

    output = {key: output[key] for key in output_vars}

    if kwargs["limit_inputs"]:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            rh_valid,
            met_valid,
            clo_valid,
        ) = check_standard_compliance_array(
            standard="fan_heatwaves", tdb=tdb, tr=tr, v=v, rh=rh, met=met, clo=clo
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        output = {key: np.where(all_valid, output[key], np.nan) for key in output_vars}

    for key in output.keys():
        # round the results if needed
        if (kwargs["round"]) and (type(output[key]) is not bool):
            output[key] = np.around(output[key], 1)

    return output



import pytest

from pythermalcomfort.models import wbgt


def test_wbgt():
    assert wbgt(25, 30) == 26.5
    assert wbgt(twb=25, tg=32) == 27.1
    assert wbgt(twb=25, tg=32, tdb=20) == 27.1
    assert wbgt(twb=25, tg=32, tdb=20, with_solar_load=True) == 25.9
    with pytest.raises(ValueError):
        wbgt(twb=25, tg=32, with_solar_load=True)
    # data from Table D.1 ISO 7243
    assert wbgt(twb=17.3, tg=40, round=True) == 24.1
    assert wbgt(twb=21.1, tg=55, round=True) == 31.3
    assert wbgt(twb=16.7, tg=40, round=True) == 23.7



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
from pythermalcomfort.models import JOS3
from pythermalcomfort.jos3_functions.utilities import local_clo_typical_ensembles

# Make "jos3_example_example" directory in the current directory
directory_name = "jos3_output_example"
current_directory = os.getcwd()
jos3_example_directory = os.path.join(current_directory, directory_name)
if not os.path.exists(jos3_example_directory):
    os.makedirs(jos3_example_directory)

# -------------------------------------------
# EXAMPLE 1 (simple simulation)
# -------------------------------------------

# Build a model and set a body built
# Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.
model = JOS3(height=1.7, weight=60, age=30)

# Set the first phase
model.to = 28  # Operative temperature [°C]
model.rh = 40  # Relative humidity [%]
model.v = 0.2  # Air velocity [m/s]
model.par = 1.2  # Physical activity ratio [-]
model.simulate(60)  # Exposure time = 60 [min]

# Set the next condition (You only need to change the parameters that you want to change)
model.to = 20  # Change only operative temperature
model.simulate(60)  # Additional exposure time = 60 [min]

# Show the results
df = pd.DataFrame(model.dict_results())  # Make pandas.DataFrame
df.t_skin_mean.plot()  # Plot time series of mean skin temperature.
plt.ylabel("Mean skin temperature [°C]")  # Set y-label as 'Mean skin temperature [°C]'
plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
plt.savefig(
    os.path.join(jos3_example_directory, "jos3_example1_mean_skin_temperature.png")
)  # Save plot at the current directory
plt.show()  # Show the plot

# Exporting the results as csv
model.to_csv(os.path.join(jos3_example_directory, "jos3_example1 (default output).csv"))

# Print the BMR value using the getter
print("BMR=", model.bmr)
print("Body name list: ", model.body_names)

# -------------------------------------------
# EXAMPLE 2 (detail simulation)
# -------------------------------------------

# Build a model and set a body built
# Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.
model = JOS3(
    height=1.7,
    weight=60,
    fat=20,
    age=30,
    sex="male",
    bmr_equation="japanese",
    bsa_equation="fujimoto",
    ex_output="all",
)

# Set environmental conditions such as air temperature, mean radiant temperature using the setter methods.
# Set the first condition
# Environmental parameters can be input as int, float, list, dict, numpy array format.
model.tdb = 28  # Air temperature [°C]
model.tr = 30  # Mean radiant temperature [°C]
model.rh = 40  # Relative humidity [%]
model.v = np.array(  # Air velocity [m/s]
    [
        0.2,  # head
        0.4,  # neck
        0.4,  # chest
        0.1,  # back
        0.1,  # pelvis
        0.4,  # left shoulder
        0.4,  # left arm
        0.4,  # left hand
        0.4,  # right shoulder
        0.4,  # right arm
        0.4,  # right hand
        0.1,  # left thigh
        0.1,  # left leg
        0.1,  # left foot
        0.1,  # right thigh
        0.1,  # right leg
        0.1,  # right foot
    ]
)
model.clo = local_clo_typical_ensembles[
    "briefs, socks, undershirt, work jacket, work pants, safety shoes"
]["local_body_part"]

# par should be input as int, float.
model.par = 1.2  # Physical activity ratio [-], assuming a sitting position
# posture should be input as int (0, 1, or 2) or str ("standing", "sitting" or "lying").
# (0="standing", 1="sitting" or 2="lying")
model.posture = "sitting"  # Posture [-], assuming a sitting position

# Run JOS-3 model
model.simulate(
    times=30,  # Number of loops of a simulation
    dtime=60,  # Time delta [sec]. The default is 60.
)  # Exposure time = 30 [loops] * 60 [sec] = 30 [min]

# Set the next condition (You only need to change the parameters that you want to change)
model.to = 20  # Change operative temperature
model.v = {  # Air velocity [m/s], assuming to use a desk fan
    "head": 0.2,
    "neck": 0.4,
    "chest": 0.4,
    "back": 0.1,
    "pelvis": 0.1,
    "left_shoulder": 0.4,
    "left_arm": 0.4,
    "left_hand": 0.4,
    "right_shoulder": 0.4,
    "right_arm": 0.4,
    "right_hand": 0.4,
    "left_thigh": 0.1,
    "left_leg": 0.1,
    "left_foot": 0.1,
    "right_thigh": 0.1,
    "right_leg": 0.1,
    "right_foot": 0.1,
}
# Run JOS-3 model
model.simulate(
    times=60,  # Number of loops of a simulation
    dtime=60,  # Time delta [sec]. The default is 60.
)  # Additional exposure time = 60 [loops] * 60 [sec] = 60 [min]

# Set the next condition (You only need to change the parameters that you want to change)
model.tdb = 30  # Change air temperature [°C]
model.tr = 35  # Change mean radiant temperature [°C]
# Run JOS-3 model
model.simulate(
    times=30,  # Number of loops of a simulation
    dtime=60,  # Time delta [sec]. The default is 60.
)  # Additional exposure time = 30 [loops] * 60 [sec] = 30 [min]

# Show the results
df = pd.DataFrame(model.dict_results())  # Make pandas.DataFrame
df[
    ["t_skin_mean", "t_skin_head", "t_skin_chest", "t_skin_left_hand"]
].plot()  # Plot time series of local skin temperature.
plt.legend(["Mean", "Head", "Chest", "Left hand"])  # Reset the legends
plt.ylabel("Skin temperature [°C]")  # Set y-label as 'Skin temperature [°C]'
plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
plt.savefig(
    os.path.join(jos3_example_directory, "jos3_example2_skin_temperatures.png")
)  # Save plot at the current directory
plt.show()  # Show the plot

# Exporting the results as csv
model.to_csv(os.path.join(jos3_example_directory, "jos3_example2 (all output).csv"))

"""
Following code is for validation between experimental and predicted data
"""
exp_dataset_name = "human_subject_experiment_dataset.xlsx"
exp_dataset_path = os.path.join(jos3_example_directory, exp_dataset_name)

# Initialize an empty dictionary to hold the datasets
exp_dataset = {}

# List of sheet names and their respective header row indices to be read
sheet_names = [("Stolwijk1966", 0), ("Werner1980", 0)]

try:
    # Loop through each sheet name and read the data into a DataFrame
    for sheet_name, header in sheet_names:
        exp_dataset[sheet_name] = pd.read_excel(
            exp_dataset_path, header=header, sheet_name=sheet_name
        )
# Handle the case where the file is not found
except FileNotFoundError:
    print(f"File {exp_dataset_path} not found.")
# Handle other general exceptions
except Exception as e:
    print(f"An error occurred: {e}")

# Concatenate all the individual data frames into a single DataFrame
sim_dataset = {}


def sim_stolwijk_hardy(models, tolist, rhlist):
    result = []
    for model in models:
        model.icl = np.array(
            [
                0,
                0,
                0,
                0.3,
                0.3,
                0,
                0,
                0,
                0,
                0,
                0,
                0.3,
                0.3,
                0,
                0.3,
                0.3,
                0,
            ]
        )
        model.par = 1.2
        model.posture = "sitting"
        # model.options["nonshivering_thermogenesis"] = True
        # model.options["limit_dshiv/dt"] = True
        # model.options["shivering_threshold"] = True

        model.to = tolist[0]
        model.rh = rhlist[0]
        model.simulate(10, dtime=600)

        model.to = tolist[1]
        model.rh = rhlist[1]
        model.simulate(60)

        model.to = tolist[2]
        model.rh = rhlist[2]
        model.simulate(120)

        model.to = tolist[3]
        model.rh = rhlist[3]
        model.simulate(60)

        sim = pd.DataFrame(model.dict_results())
        sim = sim.loc[
            10:,
            "t_skin_mean":,
        ]
        result.append(sim.copy())

    avgsim = (result[0] + result[1] + result[2]) / 3
    avgsim = avgsim.iloc[::5]
    avgsim.index = [i for i in range(0, 241, 5)]
    avgsim["time"] = [i for i in range(0, 241, 5)]

    return avgsim.copy()


simdatas = {}

# -------------------------------------------------------------------------------
# A : Stolwijk and Hardy, 1966
# J.A.J. Stolwijk, J.D. Hardy, Partitional calorimetric studies of responses of man
# to thermal transients, J. Appl. Physiol. 21 (3) (1966) 967–977,
# https://doi.org/10.1152/jappl.1966.21.6.1799.
# -------------------------------------------------------------------------------
dnames = ["A-FIG.4", "A-FIG.5", "A-FIG.6", "A-FIG.7"]

inputcons = {
    "A-FIG.4": {"to": [28.0, 27.8, 33.3, 28.0], "rh": [40, 37, 34, 37]},
    "A-FIG.5": {"to": [28.0, 28.5, 37.5, 28.5], "rh": [40, 40, 34, 41]},
    "A-FIG.6": {"to": [28.0, 28.0, 42.5, 28.1], "rh": [40, 37, 34, 37]},
    "A-FIG.7": {"to": [28.0, 28.1, 47.8, 28.3], "rh": [40, 43, 27, 44]},
}

for d in dnames:
    models = []
    models.append(
        JOS3(
            height=1.95,
            weight=88.6,
            age=25,
        )
    )
    models.append(
        JOS3(
            height=1.84,
            weight=76.1,
            age=22,
        )
    )
    models.append(
        JOS3(
            height=1.75,
            weight=110,
            age=23,
        )
    )

    tolist = inputcons[d]["to"]
    rhlist = inputcons[d]["rh"]
    output = sim_stolwijk_hardy(models, tolist, rhlist)
    print("simulated:", d)

    simdatas[d] = output

# -------------------------------------------------------------------------------
# B : Stolwijk and Hardy, 1966
#  J.D. Hardy, J.A.J. Stolwjk, Partitional exposures calorimetric studies of man
# during exposures to thermal transients, J. Appl. Physiol. 21 (1966) 1799–1806,
# https://doi.org/10.1152/jappl.1966.21.6.1799.
# -------------------------------------------------------------------------------
dnames = ["B-FIG.1", "B-FIG.2", "B-FIG.3", "B-FIG.4", "B-FIG.5"]

inputcons = {
    "B-FIG.1": {"to": [28.0, 29.0, 22.0, 29.0], "rh": [40, 44, 39, 41]},
    "B-FIG.2": {"to": [28.0, 28.0, 18.0, 28.0], "rh": [40, 40, 40, 40]},
    "B-FIG.3": {"to": [28.0, 22.3, 43.5, 22.6], "rh": [40, 40, 38, 36]},
    "B-FIG.4": {"to": [28.0, 18.0, 42.0, 18.0], "rh": [40, 40, 40, 40]},
    "B-FIG.5": {"to": [28.0, 43.0, 17.0, 43.0], "rh": [40, 40, 40, 40]},
}

for d in dnames:
    models = []
    models.append(
        JOS3(
            height=1.91,
            weight=77.2,
            age=25,
        )
    )
    models.append(
        JOS3(
            height=1.91,
            weight=84.5,
            age=26,
        )
    )
    models.append(
        JOS3(
            height=1.88,
            weight=92.7,
            age=22,
        )
    )

    tolist = inputcons[d]["to"]
    rhlist = inputcons[d]["rh"]
    output = sim_stolwijk_hardy(models, tolist, rhlist)
    print("simulated:", d)

    simdatas[d] = output

to_concat = []
for key, value in simdatas.items():
    value["Condition"] = key
    to_concat.append(value.copy())
df = pd.concat(to_concat).reset_index(drop=True)

sim_dataset["Stolwijk1966"] = df.copy()

gexp = exp_dataset["Stolwijk1966"].copy()
gsim = sim_dataset["Stolwijk1966"].copy()
dfdict = {}
for dn in [
    "A-FIG.4",
    "A-FIG.5",
    "A-FIG.6",
    "A-FIG.7",
    "B-FIG.1",
    "B-FIG.2",
    "B-FIG.3",
    "B-FIG.4",
    "B-FIG.5",
]:
    dfdict[dn] = pd.DataFrame(
        {
            "TreExp": gexp.loc[gexp["Condition"] == dn, "Tre"],
            "TreSim": gsim.loc[gsim["Condition"] == dn, "t_core_pelvis"],
            "TskExp": gexp.loc[gexp["Condition"] == dn, "Tsk"],
            "TskSim": gsim.loc[gsim["Condition"] == dn, "t_skin_mean"],
        }
    )

dfs = []
for key, value in dfdict.items():
    for seg in ["Tre", "Tsk"]:
        df = value[[seg + "Exp", seg + "Sim"]].copy()
        df.columns = ["Exp", "Sim"]
        df["Time"] = [i for i in range(0, 241, 5)]
        df["ExpName"] = key
        df["Type"] = seg
        dfs.append(df.copy())
dfstatic = pd.concat(dfs).reset_index(drop=True)


def plotdata(ser, ax, color, markersize=7, legend=False):
    mk = None
    me = 1
    if color == "black":
        ser.plot(
            ax=ax,
            linewidth=2,
            color=(0.0, 0.0, 0.0, 0.7),
            marker="o",
            markersize=markersize,
            legend=legend,
            markerfacecolor=mk,
            markeredgewidth=me,
        )
    elif color == "blue":
        ser.plot(
            ax=ax,
            linewidth=2,
            color=(0.1, 0.3, 1.0, 0.7),
            marker="^",
            markersize=markersize,
            legend=legend,
            markerfacecolor=mk,
            markeredgewidth=me,
        )
    elif color == "green":
        ser.plot(
            ax=ax,
            linewidth=2,
            color=(0.2, 0.6, 0.2, 0.7),
            marker="s",
            markersize=markersize,
            legend=legend,
            markerfacecolor=mk,
            markeredgewidth=me,
        )
    elif color == "gray":
        ser.plot(
            ax=ax,
            linewidth=2,
            color=(0.3, 0.3, 0.3, 0.7),
            marker="o",
            markersize=markersize,
            legend=legend,
            markerfacecolor=mk,
            markeredgewidth=me,
        )
    elif color == "lightblue":
        ser.plot(
            ax=ax,
            linewidth=2,
            color=(0.3, 0.1, 0.7, 0.7),
            marker="^",
            markersize=markersize,
            legend=legend,
            markerfacecolor=mk,
            markeredgewidth=me,
        )


def graph(gexp, gsim, ax):
    gexp.index = [i for i in range(0, 241, 5)]
    gsim.index = [i for i in range(0, 241, 5)]

    # Tre
    ser = gexp["Tre"]
    ser.name = "Rectum (EXP)"
    plotdata(ser, ax, "black")
    ser = gsim["t_core_pelvis"]
    ser.name = "Rectum (JOS-3)"
    plotdata(ser, ax, "blue")

    # Tsk
    ser = gexp["Tsk"]
    ser.name = "Mean skin (EXP)"
    plotdata(ser, ax, "gray")
    ser = gsim["t_skin_mean"]
    ser.name = "Mean skin (JOS-3)"
    plotdata(ser, ax, "lightblue")

    ax.set_ylim((28, 40))
    ax.set_yticks([i for i in range(28, 41, 2)])
    ax.set_xlim((0, 240))
    ax.set_xticks([i for i in range(0, 241, 30)])


inputcons = {
    "A-FIG.4": {"to": [28.0, 27.8, 33.3, 28.0], "rh": [40, 37, 34, 37]},
    "A-FIG.5": {"to": [28.0, 28.5, 37.5, 28.5], "rh": [40, 40, 34, 41]},
    "A-FIG.6": {"to": [28.0, 28.0, 42.5, 28.1], "rh": [40, 37, 34, 37]},
    "A-FIG.7": {"to": [28.0, 28.1, 47.8, 28.3], "rh": [40, 43, 27, 44]},
    "B-FIG.1": {"to": [28.0, 29.0, 22.0, 29.0], "rh": [40, 44, 39, 41]},
    "B-FIG.2": {"to": [28.0, 28.0, 18.0, 28.0], "rh": [40, 40, 40, 40]},
    "B-FIG.3": {"to": [28.0, 22.3, 43.5, 22.6], "rh": [40, 40, 38, 36]},
    "B-FIG.4": {"to": [28.0, 18.0, 42.0, 18.0], "rh": [40, 40, 40, 40]},
    "B-FIG.5": {"to": [28.0, 43.0, 17.0, 43.0], "rh": [40, 40, 40, 40]},
}

gexp = exp_dataset["Stolwijk1966"].copy()
gsim = sim_dataset["Stolwijk1966"].copy()
fig, axes = plt.subplots(nrows=5, ncols=2, sharex=True, sharey=True, figsize=(8, 12))


for i, dn in enumerate(
    [
        "A-FIG.4",
        "A-FIG.5",
        "A-FIG.6",
        "A-FIG.7",
        "B-FIG.1",
        "B-FIG.2",
        "B-FIG.3",
        "B-FIG.4",
        "B-FIG.5",
    ]
):
    # if i >= 4:
    i += 1
    yi = i % 5
    xi = i // 5
    ax = axes[yi, xi]
    graph(gexp.loc[gexp["Condition"] == dn], gsim.loc[gsim["Condition"] == dn], ax)

    ax.set_title(dn)

    ax.set_xticks([i for i in range(0, 241, 30)])
    ax.set_xticklabels([str(i) for i in range(0, 241, 30)])

    ax.axvspan(60, 180, color=(0, 0, 0, 0.1))
    ax.axhspan(39, 40, color=(1, 1, 1, 0.6))
    t = "{:.1f}°C".format(inputcons[dn]["to"][1])
    ax.text(30, 39.1, t, horizontalalignment="center", c="black", fontsize=10)
    t = "{:.1f}°C".format(inputcons[dn]["to"][2])
    ax.text(120, 39.1, t, horizontalalignment="center", c="black", fontsize=10)
    t = "{:.1f}°C".format(inputcons[dn]["to"][3])
    ax.text(210, 39.1, t, horizontalalignment="center", c="black", fontsize=10)

fig.delaxes(axes[0, 0])
handles, labels = ax.get_legend_handles_labels()
fig.legend(
    handles,
    labels,
    loc="upper left",
    bbox_to_anchor=(0.1, 0.95),
    frameon=False,
    fontsize=12,
)

fig.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.1, hspace=0.4)

fig.text(
    0.03, 0.5, "Body temperature [°C]", va="center", rotation="vertical", fontsize=12
)
fig.text(0.53, 0.03, "Time [min]", ha="center", fontsize=12)

plt.tight_layout(rect=[0.05, 0.05, 1, 1])
fig.savefig(
    os.path.join(jos3_example_directory, "jos3_validation_with_Stolwijk1966.png")
)



import numpy as np
from dataclasses import dataclass

from pythermalcomfort.psychrometrics import t_o
from pythermalcomfort.shared_functions import valid_range
from pythermalcomfort.utilities import (
    units_converter,
    check_standard_compliance_array,
)


@dataclass
class AdaptiveASHRAE:
    tmp_cmf: float
    tmp_cmf_80_low: float
    tmp_cmf_80_up: float
    tmp_cmf_90_low: float
    tmp_cmf_90_up: float
    acceptability_80: np.array
    acceptability_90: np.array

    def __getitem__(self, item):
        return getattr(self, item)


def adaptive_ashrae(
    tdb, tr, t_running_mean, v, units="SI", limit_inputs=True
) -> AdaptiveASHRAE:
    """Determines the adaptive thermal comfort based on ASHRAE 55. The adaptive
    model relates indoor design temperatures or acceptable temperature ranges
    to outdoor meteorological or climatological parameters. The adaptive model
    can only be used in occupant-controlled naturally conditioned spaces that
    meet all the following criteria:

    * There is no mechianical cooling or heating system in operation
    * Occupants have a metabolic rate between 1.0 and 1.5 met
    * Occupants are free to adapt their clothing within a range as wide as 0.5 and 1.0 clo
    * The prevailing mean (runnin mean) outdoor temperature is between 10 and 33.5 °C

    Parameters
    ----------
    tdb : float or array-like
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    tr : float or array-like
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    t_running_mean: float or array-like
        running mean temperature, default in [°C] in [°C] in [°F] if `units` = 'IP'

        The running mean temperature can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.
    v : float or array-like
        air speed, default in [m/s] in [fps] if `units` = 'IP'
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.
    limit_inputs : boolean default True
        By default, if the inputs are outsude the standard applicability limits the
        function returns nan. If False returns pmv and ppd values even if input values are
        outside the applicability limits of the model.

        The ASHRAE 55 2020 limits are 10 < tdb [°C] < 40, 10 < tr [°C] < 40,
        0 < vr [m/s] < 2, 10 < t running mean [°C] < 33.5

    Returns
    -------
    tmp_cmf : float or array-like
        Comfort temperature a that specific running mean temperature, default in [°C]
        or in [°F]
    tmp_cmf_80_low : float or array-like
        Lower acceptable comfort temperature for 80% occupants, default in [°C] or in [°F]
    tmp_cmf_80_up : float or array-like
        Upper acceptable comfort temperature for 80% occupants, default in [°C] or in [°F]
    tmp_cmf_90_low : float or array-like
        Lower acceptable comfort temperature for 90% occupants, default in [°C] or in [°F]
    tmp_cmf_90_up : float or array-like
        Upper acceptable comfort temperature for 90% occupants, default in [°C] or in [°F]
    acceptability_80 : bol or array-like
        Acceptability for 80% occupants
    acceptability_90 : bol or array-like
        Acceptability for 90% occupants

    Notes
    -----
    You can use this function to calculate if your conditions are within the `adaptive
    thermal comfort region`.
    Calculations with comply with the ASHRAE 55 2020 Standard [1]_.

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import adaptive_ashrae
        >>> results = adaptive_ashrae(tdb=25, tr=25, t_running_mean=20, v=0.1)
        >>> print(results)
        {'tmp_cmf': 24.0, 'tmp_cmf_80_low': 20.5, 'tmp_cmf_80_up': 27.5,
        'tmp_cmf_90_low': 21.5, 'tmp_cmf_90_up': 26.5, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> print(results.acceptability_80)  # or use print(results["acceptability_80"])
        True
        # The conditions you entered are considered to be comfortable for by 80% of the
        occupants

        >>> # You can also pass arrays as input to the function
        >>> results = adaptive_ashrae(tdb=[25, 26], tr=25, t_running_mean=20, v=0.1)
        >>> print(results)
        {'tmp_cmf': 24.0, 'tmp_cmf_80_low': 20.5, 'tmp_cmf_80_up': 27.5,
        'tmp_cmf_90_low': 21.5, 'tmp_cmf_90_up': 26.5, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> # for users who want to use the IP system
        >>> results = adaptive_ashrae(tdb=77, tr=77, t_running_mean=68, v=0.3, units='ip')
        >>> print(results)
        {'tmp_cmf': 75.2, 'tmp_cmf_80_low': 68.9, 'tmp_cmf_80_up': 81.5,
        'tmp_cmf_90_low': 70.7, 'tmp_cmf_90_up': 79.7, 'acceptability_80': array(True),
        'acceptability_90': array(True)}

        >>> adaptive_ashrae(tdb=25, tr=25, t_running_mean=9, v=0.1)
        {'tmp_cmf': nan, 'tmp_cmf_80_low': nan, ... }
        # The adaptive thermal comfort model can only be used
        # if the running mean temperature is higher than 10°C
    """

    tdb = np.array(tdb)
    tr = np.array(tr)
    t_running_mean = np.array(t_running_mean)
    v = np.array(v)
    standard = "ashrae"

    if units.lower() == "ip":
        tdb, tr, t_running_mean, v = units_converter(
            tdb=tdb, tr=tr, tmp_running_mean=t_running_mean, v=v
        )

    (
        tdb_valid,
        tr_valid,
        v_valid,
    ) = check_standard_compliance_array(standard, tdb=tdb, tr=tr, v=v)
    trm_valid = valid_range(t_running_mean, (10.0, 33.5))

    to = t_o(tdb, tr, v, standard=standard)

    # calculate cooling effect (ce) of elevated air speed when top > 25 degC.
    ce = np.where((v >= 0.6) & (to >= 25.0), 999, 0)
    ce = np.where((v < 0.9) & (ce == 999), 1.2, ce)
    ce = np.where((v < 1.2) & (ce == 999), 1.8, ce)
    ce = np.where(ce == 999, 2.2, ce)

    # Relation between comfort and outdoor temperature
    t_cmf = 0.31 * t_running_mean + 17.8

    if limit_inputs:
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(trm_valid)
        )
        t_cmf = np.where(all_valid, t_cmf, np.nan)

    t_cmf = np.around(t_cmf, 1)

    tmp_cmf_80_low = t_cmf - 3.5
    tmp_cmf_90_low = t_cmf - 2.5
    tmp_cmf_80_up = t_cmf + 3.5 + ce
    tmp_cmf_90_up = t_cmf + 2.5 + ce

    acceptability_80 = np.where(
        (tmp_cmf_80_low <= to) & (to <= tmp_cmf_80_up), True, False
    )
    acceptability_90 = np.where(
        (tmp_cmf_90_low <= to) & (to <= tmp_cmf_90_up), True, False
    )

    if units.lower() == "ip":
        (
            t_cmf,
            tmp_cmf_80_low,
            tmp_cmf_80_up,
            tmp_cmf_90_low,
            tmp_cmf_90_up,
        ) = units_converter(
            from_units="si",
            tmp_cmf=t_cmf,
            tmp_cmf_80_low=tmp_cmf_80_low,
            tmp_cmf_80_up=tmp_cmf_80_up,
            tmp_cmf_90_low=tmp_cmf_90_low,
            tmp_cmf_90_up=tmp_cmf_90_up,
        )

    return AdaptiveASHRAE(
        tmp_cmf=t_cmf,
        tmp_cmf_80_low=tmp_cmf_80_low,
        tmp_cmf_80_up=tmp_cmf_80_up,
        tmp_cmf_90_low=tmp_cmf_90_low,
        tmp_cmf_90_up=tmp_cmf_90_up,
        acceptability_80=acceptability_80,
        acceptability_90=acceptability_90,
    )



import csv
import datetime as dt
import os
import re

import numpy as np

from pythermalcomfort.__init__ import __version__
from pythermalcomfort.jos3_functions import construction as cons
from pythermalcomfort.jos3_functions import matrix
from pythermalcomfort.jos3_functions import thermoregulation as threg
from pythermalcomfort.jos3_functions.construction import _to17array
from pythermalcomfort.jos3_functions.matrix import (
    NUM_NODES,
    INDEX,
    VINDEX,
    BODY_NAMES,
    remove_body_name,
)
from pythermalcomfort.jos3_functions.parameters import Default, ALL_OUT_PARAMS
from pythermalcomfort.models import pmv

# # testing morris equation
# from scipy import optimize
# import matplotlib.pyplot as plt
# import numpy as np
# f, ax = plt.subplots()
# for person in ["young", "old", "meds"]:
#     t_lim = []
#     rh_array = list(range(5, 100))
#     for rh in rh_array:
#
#         def function(x):
#             return use_fans_morris(x, 3.5, rh, 70, target_person=person)
#
#         t_lim.append(optimize.brentq(function, 30, 70))
#
#     z = np.polyfit(rh_array, t_lim, 4)
#     p = np.poly1d(z)
#     y_new = p(rh_array)
#     # plt.plot(rh_array, t_lim, "o")
#     plt.plot(rh_array, y_new, "-", label=person)
# ax.set(
#     ylabel="Temperature [°C]",
#     xlabel="Relative Humidity [Rh]",
#     ylim=(24, 52),
#     xlim=(5, 70),
# )
# plt.legend()


# include also the following models:
#  radiant_tmp_asymmetry
#  draft
#  floor_surface_tmp
#  Perceived temperature (Blazejczyk2012)
#  Physiological subjective temperature and physiological strain (Blazejczyk2012)
#  more models here: https://www.rdocumentation.org/packages/comf/versions/0.1.9
#  more models here: https://rdrr.io/cran/comf/man/
#  to print the R source code use comf::pmv


class JOS3:
    """JOS-3 model simulates human thermal physiology including skin
    temperature, core temperature, sweating rate, etc. for the whole body and
    17 local body parts.

    This model was developed at Shin-ichi Tanabe Laboratory, Waseda University
    and was derived from 65 Multi-Node model (https://doi.org/10.1016/S0378-7788(02)00014-2)
    and JOS-2 model (https://doi.org/10.1016/j.buildenv.2013.04.013).

    To use this model, create an instance of the JOS3 class with optional body parameters
    such as body height, weight, age, sex, etc.

    Environmental conditions such as air temperature, mean radiant temperature, air velocity, etc.
    can be set using the setter methods. (ex. X.tdb, X.tr X.v)
    If you want to set the different conditions in each body part, set them
    as a 17 lengths of list, dictionary, or numpy array format.

    List or numpy array format input must be 17 lengths and means the order of "head", "neck", "chest",
    "back", "pelvis", "left_shoulder", "left_arm", "left_hand", "right_shoulder", "right_arm",
    "right_hand", "left_thigh", "left_leg", "left_foot", "right_thigh", "right_leg" and "right_foot".

    The model output includes local and mean skin temperature, local core temperature,
    local and mean skin wettedness, and heat loss from the skin etc.
    The model output can be accessed using "dict_results()" method and be converted to a csv file
    using "to_csv" method.
    Each output parameter also can be accessed using getter methods.
    (ex. X.t_skin, X.t_skin_mean, X.t_core)

    If you use this package, please cite us as follows and mention the version of pythermalcomfort used:
    Y. Takahashi, A. Nomoto, S. Yoda, R. Hisayama, M. Ogata, Y. Ozeki, S. Tanabe,
    Thermoregulation Model JOS-3 with New Open Source Code, Energy & Buildings (2020),
    doi: https://doi.org/10.1016/j.enbuild.2020.110575

    Note: To maintain consistency in variable names for pythermalcomfort,
    some variable names differ from those used in the original paper.
    """

    def __init__(
        self,
        height=Default.height,
        weight=Default.weight,
        fat=Default.body_fat,
        age=Default.age,
        sex=Default.sex,
        ci=Default.cardiac_index,
        bmr_equation=Default.bmr_equation,
        bsa_equation=Default.bsa_equation,
        ex_output=None,
    ):
        """Initialize a new instance of JOS3 class, which is designed to model
        and simulate various physiological parameters related to human
        thermoregulation.

        This class uses mathematical models to calculate and predict
        body temperature, basal metabolic rate, body surface area, and
        other related parameters.

        Parameters
        ----------
        height : float, optional
            body height, in [m]. The default is 1.72.
        weight : float, optional
            body weight, in [kg]. The default is 74.43.
        fat : float, optional
            fat percentage, in [%]. The default is 15.
        age : int, optional
            age, in [years]. The default is 20.
        sex : str, optional
            sex ("male" or "female"). The default is "male".
        ci : float, optional
            Cardiac index, in [L/min/m2]. The default is 2.6432.
        bmr_equation : str, optional
            The equation used to calculate basal metabolic rate (BMR). Choose a BMR equation.
            The default is "harris-benedict" equation created uding Caucasian's data. (DOI: doi.org/10.1073/pnas.4.12.370)
            If the Ganpule's equation (DOI: doi.org/10.1038/sj.ejcn.1602645) for Japanese people is used, input "japanese".
        bsa_equation : str, optional
            The equation used to calculate body surface area (bsa). Choose a bsa equation.
            You can choose "dubois", "fujimoto", "kruazumi", or "takahira". The default is "dubois".
            The body surface area can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.body_surface_area`.
        ex_output : None, list or "all", optional
            This is used when you want to display results other than the default output parameters (ex.skin temperature);
            by default, JOS outputs only the most necessary parameters in order to reduce the computational load.
            If the parameters other than the default output parameters are needed,
            specify the list of the desired parameter names in string format like ["bf_skin", "bf_core", "t_artery"].
            If you want to display all output results, set ex_output is "all".

        Attributes
        ----------
        tdb : float or array-like
            Dry bulb air temperature [°C].
        tr : float or array-like
            Mean radiant temperature [°C].
        to : float or array-like
            Operative temperature [°C].
        v : float or array-like
            Air speed [m/s].
        rh : float or array-like
            Relative humidity [%].
        clo : float or array-like
            Clothing insulation [clo].
            Note: If you want to input clothing insulation to each body part,
            it can be input using the dictionaly in "utilities.py" in "jos3_function" folder.
            :py:meth:`pythermalcomfort.jos3_functions.utilities.local_clo_typical_ensembles`.
        par : float
            Physical activity ratio [-].
            This equals the ratio of metabolic rate to basal metabolic rate.
            The par of sitting quietly is 1.2.
        posture : str
            Body posture [-]. Choose a posture from standing, sitting or lying.
        body_temp : numpy.ndarray (85,)
            All segment temperatures of JOS-3

        Methods
        -------
        simulate(times, dtime, output):
            Run JOS-3 model for given times.
        dict_results():
            Get results as a dictionary with pandas.DataFrame values.
        to_csv(path=None, folder=None, unit=True, meaning=True):
            Export results as csv format.

        Returns
        -------
        cardiac_output: cardiac output (the sum of the whole blood flow) [L/h]
        cycle_time: the counts of executing one cycle calculation [-]
        dt      : time step [sec]
        pythermalcomfort_version: version of pythermalcomfort [-]
        q_res   : heat loss by respiration [W]
        q_skin2env: total heat loss from the skin (each body part) [W]
        q_thermogenesis_total: total thermogenesis of the whole body [W]
        simulation_time: simulation times [sec]
        t_core  : core temperature (each body part) [°C]
        t_skin  : skin temperature (each body part) [°C]
        t_skin_mean: mean skin temperature [°C]
        w       : skin wettedness (each body part) [-]
        w_mean  : mean skin wettedness [-]
        weight_loss_by_evap_and_res: weight loss by the evaporation and respiration of the whole body [g/sec]
        OPTIONAL PARAMETERS : the paramters listed below are returned if ex_output = "all"
        age     : age [years]
        bf_ava_foot: AVA blood flow rate of one foot [L/h]
        bf_ava_hand: AVA blood flow rate of one hand [L/h]
        bf_core : core blood flow rate (each body part) [L/h]
        bf_fat  : fat blood flow rate (each body part) [L/h]
        bf_muscle: muscle blood flow rate (each body part) [L/h]
        bf_skin : skin blood flow rate (each body part) [L/h]
        bsa     : body surface area (each body part) [m2]
        clo     : clothing insulation (each body part) [clo]
        e_max   : maximum evaporative heat loss from the skin (each body part) [W]
        e_skin  : evaporative heat loss from the skin (each body part) [W]
        e_sweat : evaporative heat loss from the skin by only sweating (each body part) [W]
        fat     : body fat rate [%]
        height  : body height [m]
        name    : name of the model [-]
        par     : physical activity ratio [-]
        q_bmr_core: core thermogenesis by basal metabolism (each body part) [W]
        q_bmr_fat: fat thermogenesis by basal metabolism (each body part) [W]
        q_bmr_muscle: muscle thermogenesis by basal metabolism (each body part) [W]
        q_bmr_skin: skin thermogenesis by basal metabolism (each body part) [W]
        q_nst   : core thermogenesis by non-shivering (each body part) [W]
        q_res_latent: latent heat loss by respiration (each body part) [W]
        q_res_sensible: sensible heat loss by respiration (each body part) [W]
        q_shiv  : core or muscle thermogenesis by shivering (each body part) [W]
        q_skin2env_latent: latent heat loss from the skin (each body part) [W]
        q_skin2env_sensible: sensible heat loss from the skin (each body part) [W]
        q_thermogenesis_core: core total thermogenesis (each body part) [W]
        q_thermogenesis_fat: fat total thermogenesis (each body part) [W]
        q_thermogenesis_muscle: muscle total thermogenesis (each body part) [W]
        q_thermogenesis_skin: skin total thermogenesis (each body part) [W]
        q_work  : core or muscle thermogenesis by work (each body part) [W]
        r_et    : total clothing evaporative heat resistance (each body part) [(m2*kPa)/W]
        r_t     : total clothing heat resistance (each body part) [(m2*K)/W]
        rh      : relative humidity (each body part) [%]
        sex     : sex [-]
        t_artery: arterial temperature (each body part) [°C]
        t_cb    : central blood temperature [°C]
        t_core_set: core set point temperature (each body part) [°C]
        t_fat   : fat temperature (each body part) [°C]
        t_muscle: muscle temperature (each body part) [°C]
        t_skin_set: skin set point temperature (each body part) [°C]
        t_superficial_vein: superficial vein temperature (each body part) [°C]
        t_vein  : vein temperature (each body part) [°C]
        tdb     : dry bulb air temperature (each body part) [°C]
        to      : operative temperature (each body part) [°C]
        tr      : mean radiant temperature (each body part) [°C]
        v       : air velocity (each body part) [m/s]
        weight  : body weight [kg]


        Examples
        --------
        Build a model and set a body built
        Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.

        .. code-block:: python

            >>> import numpy as np
            >>> import pandas as pd
            >>> import matplotlib.pyplot as plt
            >>> import os
            >>> from pythermalcomfort.models import JOS3
            >>> from pythermalcomfort.jos3_functions.utilities import local_clo_typical_ensembles
            >>>
            >>> directory_name = "jos3_output_example"
            >>> current_directory = os.getcwd()
            >>> jos3_example_directory = os.path.join(current_directory, directory_name)
            >>> if not os.path.exists(jos3_example_directory):
            >>>     os.makedirs(jos3_example_directory)
            >>>
            >>> model = JOS3(
            >>>     height=1.7,
            >>>     weight=60,
            >>>     fat=20,
            >>>     age=30,
            >>>     sex="male",
            >>>     bmr_equation="japanese",
            >>>     bsa_equation="fujimoto",
            >>>     ex_output="all",
            >>> )
            >>> # Set environmental conditions such as air temperature, mean radiant temperature using the setter methods.
            >>> # Set the first condition
            >>> # Environmental parameters can be input as int, float, list, dict, numpy array format.
            >>> model.tdb = 28  # Air temperature [°C]
            >>> model.tr = 30  # Mean radiant temperature [°C]
            >>> model.rh = 40  # Relative humidity [%]
            >>> model.v = np.array( # Air velocity [m/s]
            >>>     [
            >>>         0.2,  # head
            >>>         0.4,  # neck
            >>>         0.4,  # chest
            >>>         0.1,  # back
            >>>         0.1,  # pelvis
            >>>         0.4,  # left shoulder
            >>>         0.4,  # left arm
            >>>         0.4,  # left hand
            >>>         0.4,  # right shoulder
            >>>         0.4,  # right arm
            >>>         0.4,  # right hand
            >>>         0.1,  # left thigh
            >>>         0.1,  # left leg
            >>>         0.1,  # left foot
            >>>         0.1,  # right thigh
            >>>         0.1,  # right leg
            >>>         0.1,  # right foot
            >>>     ]
            >>> )
            >>> model.clo = local_clo_typical_ensembles["briefs, socks, undershirt, work jacket, work pants, safety shoes"]["local_body_part"]
            >>> # par should be input as int, float.
            >>> model.par = 1.2  # Physical activity ratio [-], assuming a sitting position
            >>> # posture should be input as int (0, 1, or 2) or str ("standing", "sitting" or "lying").
            >>> # (0="standing", 1="sitting" or 2="lying")
            >>> model.posture = "sitting"  # Posture [-], assuming a sitting position
            >>>
            >>> # Run JOS-3 model
            >>> model.simulate(
            >>>     times=30,  # Number of loops of a simulation
            >>>     dtime=60,  # Time delta [sec]. The default is 60.
            >>> )  # Exposure time = 30 [loops] * 60 [sec] = 30 [min]
            >>> # Set the next condition (You only need to change the parameters that you want to change)
            >>> model.to = 20  # Change operative temperature
            >>> model.v = { # Air velocity [m/s], assuming to use a desk fan
            >>>     'head' : 0.2,
            >>>     'neck' : 0.4,
            >>>     'chest' : 0.4,
            >>>     'back': 0.1,
            >>>     'pelvis' : 0.1,
            >>>     'left_shoulder' : 0.4,
            >>>     'left_arm' : 0.4,
            >>>     'left_hand' : 0.4,
            >>>     'right_shoulder' : 0.4,
            >>>     'right_arm' : 0.4,
            >>>     'right_hand' : 0.4,
            >>>     'left_thigh' : 0.1,
            >>>     'left_leg' : 0.1,
            >>>     'left_foot' : 0.1,
            >>>     'right_thigh' : 0.1,
            >>>     'right_leg' : 0.1,
            >>>     'right_foot' : 0.1
            >>>     }
            >>> # Run JOS-3 model
            >>> model.simulate(
            >>>     times=60,  # Number of loops of a simulation
            >>>     dtime=60,  # Time delta [sec]. The default is 60.
            >>> )  # Additional exposure time = 60 [loops] * 60 [sec] = 60 [min]
            >>> # Set the next condition (You only need to change the parameters that you want to change)
            >>> model.tdb = 30  # Change air temperature [°C]
            >>> model.tr = 35  # Change mean radiant temperature [°C]
            >>> # Run JOS-3 model
            >>> model.simulate(
            >>>     times=30,  # Number of loops of a simulation
            >>>     dtime=60,  # Time delta [sec]. The default is 60.
            >>> )  # Additional exposure time = 30 [loops] * 60 [sec] = 30 [min]
            >>> # Show the results
            >>> df = pd.DataFrame(model.dict_results())  # Make pandas.DataFrame
            >>> df[["t_skin_mean", "t_skin_head", "t_skin_chest", "t_skin_left_hand"]].plot()  # Plot time series of local skin temperature.
            >>> plt.legend(["Mean", "Head", "Chest", "Left hand"])  # Reset the legends
            >>> plt.ylabel("Skin temperature [°C]")  # Set y-label as 'Skin temperature [°C]'
            >>> plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
            >>> plt.savefig(os.path.join(jos3_example_directory, "jos3_example2_skin_temperatures.png"))  # Save plot at the current directory
            >>> plt.show()  # Show the plot
            >>> # Exporting the results as csv
            >>> model.to_csv(os.path.join(jos3_example_directory, "jos3_example2 (all output).csv"))
        """

        # Version of pythermalcomfort
        version_string = (
            __version__  # get the current version of pythermalcomfort package
        )
        version_number_string = re.findall("\d+\.\d+\.\d+", version_string)[0]
        self._version = version_number_string  # (ex. 'X.Y.Z')

        # Initialize basic attributes
        self._height = height
        self._weight = weight
        self._fat = fat
        self._sex = sex
        self._age = age
        self._ci = ci
        self._bmr_equation = bmr_equation
        self._bsa_equation = bsa_equation
        self._ex_output = ex_output

        # Calculate body surface area (bsa) rate
        self._bsa_rate = cons.bsa_rate(height, weight, bsa_equation)

        # Calculate local body surface area
        self._bsa = cons.local_bsa(height, weight, bsa_equation)

        # Calculate basal blood flow (BFB) rate [-]
        self._bfb_rate = cons.bfb_rate(height, weight, bsa_equation, age, ci)

        # Calculate thermal conductance (CDT) [W/K]
        self._cdt = cons.conductance(height, weight, bsa_equation, fat)

        # Calculate thermal capacity [J/K]
        self._cap = cons.capacity(height, weight, bsa_equation, age, ci)

        # Set initial core and skin temperature set points [°C]
        self.setpt_cr = np.ones(17) * Default.core_temperature
        self.setpt_sk = np.ones(17) * Default.skin_temperature

        # Initialize body temperature [°C]
        self._bodytemp = np.ones(NUM_NODES) * Default.other_body_temperature

        # Initialize environmental conditions and other factors
        # (Default values of input conditions)
        self._ta = np.ones(17) * Default.dry_bulb_air_temperature
        self._tr = np.ones(17) * Default.mean_radiant_temperature
        self._rh = np.ones(17) * Default.relative_humidity
        self._va = np.ones(17) * Default.air_speed
        self._clo = np.ones(17) * Default.clothing_insulation
        self._iclo = np.ones(17) * Default.clothing_vapor_permeation_efficiency
        self._par = Default.physical_activity_ratio
        self._posture = Default.posture
        self._hc = None  # Convective heat transfer coefficient
        self._hr = None  # Radiative heat transfer coefficient
        self.ex_q = np.zeros(NUM_NODES)  # External heat gain
        self._t = dt.timedelta(0)  # Elapsed time
        self._cycle = 0  # Cycle time
        self.model_name = "JOS3"  # Model name
        self.options = {
            "nonshivering_thermogenesis": True,
            "cold_acclimated": False,
            "shivering_threshold": False,
            "limit_dshiv/dt": False,
            "bat_positive": False,
            "ava_zero": False,
            "shivering": False,
        }

        # Set shivering threshold = 0
        threg.PRE_SHIV = 0

        # Initialize history to store model parameters
        self._history = []

        # Set elapsed time and cycle time to 0
        self._t = dt.timedelta(0)  # Elapsed time
        self._cycle = 0  # Cycle time

        # Reset set-point temperature and save the last model parameters
        dictout = self._reset_setpt(par=Default.physical_activity_ratio)
        self._history.append(dictout)

    def _calculate_operative_temp_when_pmv_is_zero(
        self,
        va=Default.air_speed,
        rh=Default.relative_humidity,
        met=Default.metabolic_rate,
        clo=Default.clothing_insulation,
    ) -> float:
        """Calculate operative temperature [°C] when PMV=0.

        Parameters
        ----------
        va : float, optional
            Air velocity [m/s]. The default is 0.1.
        rh : float, optional
            Relative humidity [%]. The default is 50.
        met : float, optional
            Metabolic rate [met]. The default is 1.
        clo : float, optional
            Clothing insulation [clo]. The default is 0.

        Returns
        -------
        to : float
            Operative temperature [°C].
        """

        to = 28  # initial operative temperature
        # Iterate until the PMV (Predicted Mean Vote) value is less than 0.001
        for _ in range(100):
            vpmv = pmv(tdb=to, tr=to, vr=va, rh=rh, met=met, clo=clo)
            # Break the loop if the absolute value of PMV is less than 0.001
            if abs(vpmv) < 0.001:
                break
            # Update the temperature based on the PMV value
            else:
                to = to - vpmv / 3
        return to

    def _reset_setpt(self, par=Default.physical_activity_ratio):
        """Reset set-point temperatures under steady state calculation.
        Set-point temperatures are hypothetical core or skin temperatures in a thermally neutral state
        when at rest (similar to room set-point temperature for air conditioning).
        This function is used during initialization to calculate the set-point temperatures as a reference for thermoregulation.
        Be careful, input parameters (tdb, tr, rh, v, clo, par) and body temperatures are also reset.

        Returns
        -------
        dict
            Parameters of JOS-3 model.
        """
        # Set operative temperature under PMV=0 environment
        # 1 met = 58.15 W/m2
        w_per_m2_to_met = 1 / 58.15  # unit converter W/m2 to met
        met = self.bmr * par * w_per_m2_to_met  # [met]
        self.to = self._calculate_operative_temp_when_pmv_is_zero(met=met)
        self.rh = Default.relative_humidity
        self.v = Default.air_speed
        self.clo = Default.clothing_insulation
        self.par = par  # Physical activity ratio

        # Steady-calculation
        self.options["ava_zero"] = True
        for _ in range(10):
            dict_out = self._run(dtime=60000, passive=True)

        # Set new set-point temperatures for core and skin
        self.setpt_cr = self.t_core
        self.setpt_sk = self.t_skin
        self.options["ava_zero"] = False

        return dict_out

    def simulate(self, times, dtime=60, output=True):
        """Run JOS-3 model.

        Parameters
        ----------
        times : int
            Number of loops of a simulation.
        dtime : int or float, optional
            Time delta in seconds. The default is 60.
        output : bool, optional
            If you don't want to record parameters, set False. The default is True.

        Returns
        -------
        None.
        """
        # Loop through the simulation for the given number of times
        for _ in range(times):
            # Increment the elapsed time by the time delta
            self._t += dt.timedelta(0, dtime)

            # Increment the cycle counter
            self._cycle += 1

            # Execute the simulation step
            dict_data = self._run(dtime=dtime, output=output)

            # If output is True, append the results to the history
            if output:
                self._history.append(dict_data)

    def _run(self, dtime=60, passive=False, output=True):
        """Runs the model once and gets the model parameters.

        The function then calculates several thermoregulation parameters using the input data,
        such as convective and radiative heat transfer coefficients, operative temperature, heat resistance,
        and blood flow rates.

        It also calculates the thermogenesis by shivering and non-shivering, basal thermogenesis, and thermogenesis by work.

        The function then calculates the total heat loss and gains, including respiratory,
        sweating, and extra heat gain, and builds the matrices required
        to solve for the new body temperature.

        It then calculates the new body temperature by solving the matrices using numpy's linalg library.

        Finally, the function returns a dictionary of the simulation results.
        The output parameters include cycle time, model time, t_skin_mean, t_skin, t_core, w_mean, w, weight_loss_by_evap_and_res, cardiac_output, q_thermogenesis_total, q_res, and q_skin_env.

        Additionally, if the _ex_output variable is set to "all" or is a list of keys,
        the function also returns a detailed dictionary of all the thermoregulation parameters
        and other variables used in the simulation.

        Parameters
        ----------
        dtime : int or float, optional
            Time delta [sec]. Default is 60.
        passive : bool, optional
            If you run a passive model, set to True. Default is False.
        output : bool, optional
            If you don't need parameters, set to False. Default is True.

        Returns
        -------
        dict_out : dictionary
            Output parameters.
        """

        # Compute convective and radiative heat transfer coefficient [W/(m2*K)]
        # based on posture, air velocity, air temperature, and skin temperature.
        # Manual setting is possible by setting self._hc and self._hr.
        # Compute heat and evaporative heat resistance [m2.K/W], [m2.kPa/W]

        # Get core and skin temperatures
        tcr = self.t_core
        tsk = self.t_skin

        # Convective and radiative heat transfer coefficients [W/(m2*K)]
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._va,
                self._ta,
                tsk,
            ),
            self._va,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            )
        )

        # Manually set convective and radiative heat transfer coefficients if necessary
        if self._hc is not None:
            hc = self._hc
        if self._hr is not None:
            hr = self._hr

        # Compute operative temp. [°C], clothing heat and evaporative resistance [m2.K/W], [m2.kPa/W]
        # Operative temp. [°C]
        to = threg.operative_temp(
            self._ta,
            self._tr,
            hc,
            hr,
        )
        # Clothing heat resistance [m2.K/W]
        r_t = threg.dry_r(hc, hr, self._clo)
        # Clothing evaporative resistance [m2.kPa/W]
        r_et = threg.wet_r(hc, self._clo, self._iclo)

        # ------------------------------------------------------------------
        # Thermoregulation
        # 1) Sweating
        # 2) Vasoconstriction, Vasodilation
        # 3) Shivering and non-shivering thermogenesis
        # ------------------------------------------------------------------

        # Compute the difference between the set-point temperature and body temperatures
        # and other thermoregulation parameters.
        # If running a passive model, the set-point temperature of thermoregulation is
        # set to the current body temperature.

        # set-point temperature for thermoregulation
        if passive:
            setpt_cr = tcr.copy()
            setpt_sk = tsk.copy()
        else:
            setpt_cr = self.setpt_cr.copy()
            setpt_sk = self.setpt_sk.copy()

        # Error signal = Difference between set-point and body temperatures
        err_cr = tcr - setpt_cr
        err_sk = tsk - setpt_sk

        # SWEATING THERMOREGULATION
        # Skin wettedness [-], e_skin, e_max, e_sweat [W]
        # Calculate skin wettedness, sweating heat loss, maximum sweating rate, and total sweat rate
        wet, e_sk, e_max, e_sweat = threg.evaporation(
            err_cr,
            err_sk,
            tsk,
            self._ta,
            self._rh,
            r_et,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
        )

        # VASOCONSTRICTION, VASODILATION
        # Calculate skin blood flow and basal skin blood flow [L/h]
        bf_skin = threg.skin_blood_flow(
            err_cr,
            err_sk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )

        # Calculate hands and feet's AVA blood flow [L/h]
        bf_ava_hand, bf_ava_foot = threg.ava_blood_flow(
            err_cr,
            err_sk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )
        if self.options["ava_zero"] and passive:
            bf_ava_hand = 0
            bf_ava_foot = 0

        # SHIVERING AND NON-SHIVERING
        # Calculate shivering thermogenesis [W]
        q_shiv = threg.shivering(
            err_cr,
            err_sk,
            tcr,
            tsk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._sex,
            dtime,
            self.options,
        )

        # Calculate non-shivering thermogenesis (NST) [W]
        if self.options["nonshivering_thermogenesis"]:
            q_nst = threg.nonshivering(
                err_sk,
                self._height,
                self._weight,
                self._bsa_equation,
                self._age,
                self.options["cold_acclimated"],
                self.options["bat_positive"],
            )
        else:  # not consider NST
            q_nst = np.zeros(17)

        # ------------------------------------------------------------------
        # Thermogenesis
        # ------------------------------------------------------------------

        # Calculate local basal metabolic rate (BMR) [W]
        q_bmr_local = threg.local_mbase(
            self._height,
            self._weight,
            self._age,
            self._sex,
            self._bmr_equation,
        )
        # Calculate overall basal metabolic rate (BMR) [W]
        q_bmr_total = sum([m.sum() for m in q_bmr_local])

        # Calculate thermogenesis by work [W]
        q_work = threg.local_q_work(q_bmr_total, self._par)

        # Calculate the sum of thermogenesis in core, muscle, fat, skin [W]
        (
            q_thermogenesis_core,
            q_thermogenesis_muscle,
            q_thermogenesis_fat,
            q_thermogenesis_skin,
        ) = threg.sum_m(
            q_bmr_local,
            q_work,
            q_shiv,
            q_nst,
        )
        q_thermogenesis_total = (
            q_thermogenesis_core.sum()
            + q_thermogenesis_muscle.sum()
            + q_thermogenesis_fat.sum()
            + q_thermogenesis_skin.sum()
        )

        # ------------------------------------------------------------------
        # Others
        # ------------------------------------------------------------------
        # Calculate blood flow in core, muscle, fat [L/h]
        bf_core, bf_muscle, bf_fat = threg.cr_ms_fat_blood_flow(
            q_work,
            q_shiv,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )

        # Calculate heat loss by respiratory
        p_a = threg.antoine(self._ta) * self._rh / 100
        res_sh, res_lh = threg.resp_heat_loss(
            self._ta[0], p_a[0], q_thermogenesis_total
        )

        # Calculate sensible heat loss [W]
        shl_sk = (tsk - to) / r_t * self._bsa

        # Calculate cardiac output [L/h]
        co = threg.sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)

        # Calculate weight loss rate by evaporation [g/sec]
        wlesk = (e_sweat + 0.06 * e_max) / 2418
        wleres = res_lh / 2418

        # ------------------------------------------------------------------
        # Matrix
        # This code section is focused on constructing and calculating
        # various matrices required for modeling the thermoregulation
        # of the human body.
        # Since JOS-3 has 85 thermal nodes, the determinant of 85*85 is to be solved.
        # ------------------------------------------------------------------

        # Matrix A = Matrix for heat exchange due to blood flow and conduction occurring between tissues
        # (85, 85,) ndarray

        # Calculates the blood flow in arteries and veins for core, muscle, fat, skin,
        # and arteriovenous anastomoses (AVA) in hands and feet,
        # and combines them into two arrays:
        # 1) bf_local for the local blood flow and 2) bf_whole for the whole-body blood flow.
        # These arrays are then combined to form arr_bf.
        bf_art, bf_vein = matrix.vessel_blood_flow(
            bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot
        )
        bf_local = matrix.local_arr(
            bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot
        )
        bf_whole = matrix.whole_body(bf_art, bf_vein, bf_ava_hand, bf_ava_foot)
        arr_bf = np.zeros((NUM_NODES, NUM_NODES))
        arr_bf += bf_local
        arr_bf += bf_whole

        # Adjusts the units of arr_bf from [W/K] to [/sec] and then to [-]
        # by dividing by the heat capacity self._cap and multiplying by the time step dtime.
        arr_bf /= self._cap.reshape((NUM_NODES, 1))  # Change unit [W/K] to [/sec]
        arr_bf *= dtime  # Change unit [/sec] to [-]

        # Performs similar unit conversions for the convective heat transfer coefficient array arr_cdt
        # (also divided by self._cap and multiplied by dtime).
        arr_cdt = self._cdt.copy()
        arr_cdt /= self._cap.reshape((NUM_NODES, 1))  # Change unit [W/K] to [/sec]
        arr_cdt *= dtime  # Change unit [/sec] to [-]

        # Matrix B = Matrix for heat transfer between skin and environment
        arr_b = np.zeros(NUM_NODES)
        arr_b[INDEX["skin"]] += 1 / r_t * self._bsa
        arr_b /= self._cap  # Change unit [W/K] to [/sec]
        arr_b *= dtime  # Change unit [/sec] to [-]

        # Calculates the off-diagonal and diagonal elements of the matrix A,
        # which represents the heat transfer coefficients between different parts of the body,
        # and combines them to form the full matrix A (arrA).
        # Then, the inverse of matrix A is computed (arrA_inv).
        arr_a_tria = -(arr_cdt + arr_bf)

        arr_a_dia = arr_cdt + arr_bf
        arr_a_dia = arr_a_dia.sum(axis=1) + arr_b
        arr_a_dia = np.diag(arr_a_dia)
        arr_a_dia += np.eye(NUM_NODES)

        arr_a = arr_a_tria + arr_a_dia
        arr_a_inv = np.linalg.inv(arr_a)

        # Matrix Q = Matrix for heat generation rate from thermogenesis, respiratory, sweating,
        # and extra heat gain processes in different body parts.

        # Matrix Q [W] / [J/K] * [sec] = [-]
        # Thermogensis
        arr_q = np.zeros(NUM_NODES)
        arr_q[INDEX["core"]] += q_thermogenesis_core
        arr_q[INDEX["muscle"]] += q_thermogenesis_muscle[VINDEX["muscle"]]
        arr_q[INDEX["fat"]] += q_thermogenesis_fat[VINDEX["fat"]]
        arr_q[INDEX["skin"]] += q_thermogenesis_skin

        # Respiratory [W]
        arr_q[INDEX["core"][2]] -= res_sh + res_lh  # chest core

        # Sweating [W]
        arr_q[INDEX["skin"]] -= e_sk

        # Extra heat gain [W]
        arr_q += self.ex_q.copy()

        arr_q /= self._cap  # Change unit [W]/[J/K] to [K/sec]
        arr_q *= dtime  # Change unit [K/sec] to [K]

        # Boundary batrix [°C]
        arr_to = np.zeros(NUM_NODES)
        arr_to[INDEX["skin"]] += to

        # Combines the current body temperature, the boundary matrix, and the heat generation matrix
        # to calculate the new body temperature distribution (arr).
        arr = self._bodytemp + arr_b * arr_to + arr_q

        # ------------------------------------------------------------------
        # New body temp. [°C]
        # ------------------------------------------------------------------
        self._bodytemp = np.dot(arr_a_inv, arr)

        # ------------------------------------------------------------------
        # Output parameters
        # ------------------------------------------------------------------
        dict_out = {}
        if output:  # Default output
            dict_out["pythermalcomfort_version"] = self._version
            dict_out["cycle_time"] = self._cycle
            dict_out["simulation_time"] = self._t
            dict_out["dt"] = dtime
            dict_out["t_skin_mean"] = self.t_skin_mean
            dict_out["t_skin"] = self.t_skin
            dict_out["t_core"] = self.t_core
            dict_out["w_mean"] = np.average(wet, weights=Default.local_bsa)
            dict_out["w"] = wet
            dict_out["weight_loss_by_evap_and_res"] = wlesk.sum() + wleres
            dict_out["cardiac_output"] = co
            dict_out["q_thermogenesis_total"] = q_thermogenesis_total
            dict_out["q_res"] = res_sh + res_lh
            dict_out["q_skin2env"] = shl_sk + e_sk

        detail_out = {}
        if self._ex_output and output:
            detail_out["name"] = self.model_name
            detail_out["height"] = self._height
            detail_out["weight"] = self._weight
            detail_out["bsa"] = self._bsa
            detail_out["fat"] = self._fat
            detail_out["sex"] = self._sex
            detail_out["age"] = self._age
            detail_out["t_core_set"] = setpt_cr
            detail_out["t_skin_set"] = setpt_sk
            detail_out["t_cb"] = self.t_cb
            detail_out["t_artery"] = self.t_artery
            detail_out["t_vein"] = self.t_vein
            detail_out["t_superficial_vein"] = self.t_superficial_vein
            detail_out["t_muscle"] = self.t_muscle
            detail_out["t_fat"] = self.t_fat
            detail_out["to"] = to
            detail_out["r_t"] = r_t
            detail_out["r_et"] = r_et
            detail_out["tdb"] = self._ta.copy()
            detail_out["tr"] = self._tr.copy()
            detail_out["rh"] = self._rh.copy()
            detail_out["v"] = self._va.copy()
            detail_out["par"] = self._par
            detail_out["clo"] = self._clo.copy()
            detail_out["e_skin"] = e_sk
            detail_out["e_max"] = e_max
            detail_out["e_sweat"] = e_sweat
            detail_out["bf_core"] = bf_core
            detail_out["bf_muscle"] = bf_muscle[VINDEX["muscle"]]
            detail_out["bf_fat"] = bf_fat[VINDEX["fat"]]
            detail_out["bf_skin"] = bf_skin
            detail_out["bf_ava_hand"] = bf_ava_hand
            detail_out["bf_ava_foot"] = bf_ava_foot
            detail_out["q_bmr_core"] = q_bmr_local[0]
            detail_out["q_bmr_muscle"] = q_bmr_local[1][VINDEX["muscle"]]
            detail_out["q_bmr_fat"] = q_bmr_local[2][VINDEX["fat"]]
            detail_out["q_bmr_skin"] = q_bmr_local[3]
            detail_out["q_work"] = q_work
            detail_out["q_shiv"] = q_shiv
            detail_out["q_nst"] = q_nst
            detail_out["q_thermogenesis_core"] = q_thermogenesis_core
            detail_out["q_thermogenesis_muscle"] = q_thermogenesis_muscle[
                VINDEX["muscle"]
            ]
            detail_out["q_thermogenesis_fat"] = q_thermogenesis_fat[VINDEX["fat"]]
            detail_out["q_thermogenesis_skin"] = q_thermogenesis_skin
            dict_out["q_skin2env_sensible"] = shl_sk
            dict_out["q_skin2env_latent"] = e_sk
            dict_out["q_res_sensible"] = res_sh
            dict_out["q_res_latent"] = res_lh

        if self._ex_output == "all":
            dict_out.update(detail_out)
        elif isinstance(self._ex_output, list):  # if ex_out type is list
            out_keys = detail_out.keys()
            for key in self._ex_output:
                if key in out_keys:
                    dict_out[key] = detail_out[key]
        return dict_out

    def dict_results(self):
        """Get results as a dictionary with pandas.DataFrame values.

        Returns
        -------
        dict
            A dictionary of the results, with keys as column names and values as pandas.DataFrame objects.
        """
        if not self._history:
            print("The model has no data.")
            return None

        def check_word_contain(word, *args):
            """Check if word contains *args."""
            bool_filter = False
            for arg in args:
                if arg in word:
                    bool_filter = True
            return bool_filter

        # Set column titles
        # If the values are iter, add the body names as suffix words.
        # If the values are not iter and the single value data, convert it to iter.
        key2keys = {}  # Column keys
        for key, value in self._history[0].items():
            try:
                length = len(value)
                if isinstance(value, str):
                    keys = [key]  # str is iter. Convert to list without suffix
                elif check_word_contain(key, "sve", "sfv", "superficialvein"):
                    keys = [key + "_" + BODY_NAMES[i] for i in VINDEX["sfvein"]]
                elif check_word_contain(key, "ms", "muscle"):
                    keys = [key + "_" + BODY_NAMES[i] for i in VINDEX["muscle"]]
                elif check_word_contain(key, "fat"):
                    keys = [key + "_" + BODY_NAMES[i] for i in VINDEX["fat"]]
                elif length == 17:  # if data contains 17 values
                    keys = [key + "_" + bn for bn in BODY_NAMES]
                else:
                    keys = [key + "_" + BODY_NAMES[i] for i in range(length)]
            except TypeError:  # if the value is not iter.
                keys = [key]  # convert to iter
            key2keys.update({key: keys})

        data = []
        for i, dictout in enumerate(self._history):
            row = {}
            for key, value in dictout.items():
                keys = key2keys[key]
                if len(keys) == 1:
                    values = [value]  # make list if value is not iter
                else:
                    values = value
                row.update(dict(zip(keys, values)))
            data.append(row)

        out_dict = dict(zip(data[0].keys(), [[] for _ in range(len(data[0].keys()))]))
        for row in data:
            for k in data[0].keys():
                out_dict[k].append(row[k])
        return out_dict

    def to_csv(
        self,
        path: str = None,
        folder: str = None,
        unit: bool = True,
        meaning: bool = True,
    ) -> None:
        """Export results as csv format.

        Parameters
        ----------
        path : str, optional
            Output path. If you don't use the default file name, set a name.
            The default is None.
        folder : str, optional
            Output folder. If you use the default file name with the current time,
            set a only folder path.
            The default is None.
        unit : bool, optional
            Write units in csv file. The default is True.
        meaning : bool, optional
            Write meanings of the parameters in csv file. The default is True.

        Returns
        -------
        None

        Examples
        ----------
        >>> from pythermalcomfort.models import JOS3
        >>> model = JOS3()
        >>> model.simulate(60)
        >>> model.to_csv()
        """

        # Use the model name and current time as default output file name
        if path is None:
            now_time = dt.datetime.now().strftime("%Y%m%d-%H%M%S")
            path = "{}_{}.csv".format(self.model_name, now_time)
            if folder:
                os.makedirs(folder, exist_ok=True)
                path = folder + os.sep + path
        elif not ((path[-4:] == ".csv") or (path[-4:] == ".txt")):
            path += ".csv"

        # Get simulation results as a dictionary
        dict_out = self.dict_results()

        # Get column names, units and meanings
        columns = [k for k in dict_out.keys()]
        units = []
        meanings = []
        for col in columns:
            param, body_name = remove_body_name(col)
            if param in ALL_OUT_PARAMS:
                u = ALL_OUT_PARAMS[param]["unit"]
                units.append(u)

                m = ALL_OUT_PARAMS[param]["meaning"]
                if body_name:
                    # Replace underscores with spaces
                    body_name_with_spaces = body_name.replace("_", " ")
                    meanings.append(m.replace("each body part", body_name_with_spaces))
                else:
                    meanings.append(m)
            else:
                units.append("")
                meanings.append("")

        # Write to csv file
        with open(path, "wt", newline="", encoding="utf-8-sig") as f:
            writer = csv.writer(f)
            writer.writerow(list(columns))
            if unit:
                writer.writerow(units)
            if meaning:
                writer.writerow(meanings)
            for i in range(len(dict_out["cycle_time"])):
                row = []
                for k in columns:
                    row.append(dict_out[k][i])
                writer.writerow(row)

    def _set_ex_q(self, tissue, value):
        """Set extra heat gain by tissue name.

        Parameters
        ----------
        tissue : str
            Tissue name. "core", "skin", or "artery".... If you set value to
            head muscle and other segment's core, set "all_muscle".
        value : int, float, array
            Heat gain [W]

        Returns
        -------
        array
            Extra heat gain of model.
        """
        self.ex_q[INDEX[tissue]] = value
        return self.ex_q

    @property
    def tdb(self):
        """Dry-bulb air temperature. The setter accepts int, float, dict, list,
        ndarray. The inputs are used to create a 17-element array. dict should
        be passed with BODY_NAMES as keys.

        Returns
        -------
        ndarray
            A NumPy array of shape (17,).
        """
        return self._ta

    @tdb.setter
    def tdb(self, inp):
        self._ta = _to17array(inp)

    @property
    def tr(self):
        """tr : numpy.ndarray (17,) Mean radiant temperature [°C]."""
        return self._tr

    @tr.setter
    def tr(self, inp):
        self._tr = _to17array(inp)

    @property
    def to(self):
        """to : numpy.ndarray (17,) Operative temperature [°C]."""
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._va,
                self._ta,
                self.t_skin,
            ),
            self._va,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            )
        )
        to = threg.operative_temp(
            self._ta,
            self._tr,
            hc,
            hr,
        )
        return to

    @to.setter
    def to(self, inp):
        self._ta = _to17array(inp)
        self._tr = _to17array(inp)

    @property
    def rh(self):
        """rh : numpy.ndarray (17,) Relative humidity [%]."""
        return self._rh

    @rh.setter
    def rh(self, inp):
        self._rh = _to17array(inp)

    @property
    def v(self):
        """v : numpy.ndarray (17,) Air velocity [m/s]."""
        return self._va

    @v.setter
    def v(self, inp):
        self._va = _to17array(inp)

    @property
    def posture(self):
        """posture : str Current JOS3 posture."""
        return self._posture

    @posture.setter
    def posture(self, inp):
        if inp == 0:
            self._posture = "standing"
        elif inp == 1:
            self._posture = "sitting"
        elif inp == 2:
            self._posture = "lying"
        elif type(inp) == str:
            if inp.lower() == "standing":
                self._posture = "standing"
            elif inp.lower() in ["sitting", "sedentary"]:
                self._posture = "sitting"
            elif inp.lower() in ["lying", "supine"]:
                self._posture = "lying"
        else:
            self._posture = "standing"
            print('posture must be 0="standing", 1="sitting" or 2="lying".')
            print('posture was set "standing".')

    @property
    def clo(self):
        """clo : numpy.ndarray (17,) Clothing insulation [clo]."""
        return self._clo

    @clo.setter
    def clo(self, inp):
        self._clo = _to17array(inp)

    @property
    def par(self):
        """par : float Physical activity ratio [-].This equals the ratio of metabolic rate to basal metabolic rate. par of sitting quietly is 1.2."""
        return self._par

    @par.setter
    def par(self, inp):
        self._par = inp

    @property
    def body_temp(self):
        """body_temp : numpy.ndarray (85,) All segment temperatures of JOS-3"""
        return self._bodytemp

    @body_temp.setter
    def body_temp(self, inp):
        self._bodytemp = inp.copy()

    @property
    def bsa(self):
        """bsa : numpy.ndarray (17,) Body surface areas by local body segments [m2]."""
        return self._bsa.copy()

    @property
    def r_t(self):
        """r_t : numpy.ndarray (17,) Dry heat resistances between the skin and ambience areas by local body segments [(m2*K)/W]."""
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._va,
                self._ta,
                self.t_skin,
            ),
            self._va,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            )
        )
        return threg.dry_r(hc, hr, self._clo)

    @property
    def r_et(self):
        """r_et : numpy.ndarray (17,) w (Evaporative) heat resistances between the skin and ambience areas by local body segments [(m2*kPa)/W]."""
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._va,
                self._ta,
                self.t_skin,
            ),
            self._va,
        )
        return threg.wet_r(hc, self._clo, self._iclo)

    @property
    def w(self):
        """w : numpy.ndarray (17,) Skin wettedness on local body segments [-]."""
        err_cr = self.t_core - self.setpt_cr
        err_sk = self.t_skin - self.setpt_sk
        wet, *_ = threg.evaporation(
            err_cr,
            err_sk,
            self.t_skin,
            self._ta,
            self._rh,
            self.r_et,
            self._bsa_rate,
            self._age,
        )
        return wet

    @property
    def w_mean(self):
        """w_mean : float Mean skin wettedness of the whole body [-]."""
        wet = self.w
        return np.average(wet, weights=Default.local_bsa)

    @property
    def t_skin_mean(self):
        """t_skin_mean : float Mean skin temperature of the whole body [°C]."""
        return np.average(self._bodytemp[INDEX["skin"]], weights=Default.local_bsa)

    @property
    def t_skin(self):
        """t_skin : numpy.ndarray (17,) Skin temperatures by the local body segments [°C]."""
        return self._bodytemp[INDEX["skin"]].copy()

    @t_skin.setter
    def t_skin(self, inp):
        self._bodytemp[INDEX["skin"]] = _to17array(inp)

    @property
    def t_core(self):
        """t_core : numpy.ndarray (17,) Skin temperatures by the local body segments [°C]."""
        return self._bodytemp[INDEX["core"]].copy()

    @property
    def t_cb(self):
        """t_cb : numpy.ndarray (1,) Temperature at central blood pool [°C]."""
        return self._bodytemp[0].copy()

    @property
    def t_artery(self):
        """t_artery : numpy.ndarray (17,) Arterial temperatures by the local body segments [°C]."""
        return self._bodytemp[INDEX["artery"]].copy()

    @property
    def t_vein(self):
        """t_vein : numpy.ndarray (17,) Vein temperatures by the local body segments [°C]."""
        return self._bodytemp[INDEX["vein"]].copy()

    @property
    def t_superficial_vein(self):
        """t_superficial_vein : numpy.ndarray (12,) Superficial vein temperatures by the local body segments [°C]."""
        return self._bodytemp[INDEX["sfvein"]].copy()

    @property
    def t_muscle(self):
        """t_muscle : numpy.ndarray (2,) Muscle temperatures of head and pelvis [°C]."""
        return self._bodytemp[INDEX["muscle"]].copy()

    @property
    def t_fat(self):
        """t_fat : numpy.ndarray (2,) fat temperatures of head and pelvis  [°C]."""
        return self._bodytemp[INDEX["fat"]].copy()

    @property
    def body_names(self):
        """body_names : list JOS3 body names"""
        return BODY_NAMES

    @property
    def results(self):
        """Results of the model: dict."""
        return self.dict_results()

    @property
    def bmr(self):
        """bmr : float Basal metabolic rate [W/m2]."""
        tcr = threg.basal_met(
            self._height,
            self._weight,
            self._age,
            self._sex,
            self._bmr_equation,
        )
        return tcr / self.bsa.sum()

    @property
    def version(self):
        """version : float The current version of pythermalcomfort."""
        return self._version



import numpy as np

from pythermalcomfort.models import (
    use_fans_heatwaves,
)


def test_use_fans_heatwaves():
    # checking that returns np.nan when outside standard applicability limits
    np.testing.assert_equal(
        use_fans_heatwaves(
            tdb=[41, 60],
            tr=40,
            v=0.1,
            rh=50,
            met=1.1,
            clo=0.5,
        )["e_skin"],
        [65.2, np.nan],
    )

    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=0.7, clo=0.3, body_position="sitting"
        )["heat_strain_w"]
        == False
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=0.7, clo=0.5, body_position="sitting"
        )["q_skin"]
        == 37.6
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=0.7, clo=0.7, body_position="sitting"
        )["m_rsw"]
        == 68.6
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=1.3, clo=0.3, body_position="sitting"
        )["m_rsw"]
        == 118.5
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=1.3, clo=0.5, body_position="sitting"
        )["m_rsw"]
        == 117.3
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=1.3, clo=0.7, body_position="sitting"
        )["m_rsw"]
        == 116.4
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=2, clo=0.3, body_position="sitting"
        )["heat_strain_w"]
        == False
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=2, clo=0.5, body_position="sitting"
        )["w"]
        == 0.5
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=20, met=2, clo=0.7, body_position="sitting"
        )["t_skin"]
        == 36.2
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=40, met=0.7, clo=0.3, body_position="sitting"
        )["heat_strain_blood_flow"]
        == False
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=40, met=0.7, clo=0.5, body_position="sitting"
        )["t_core"]
        == 36.9
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=40, met=0.7, clo=0.7, body_position="sitting"
        )["m_rsw"]
        == 73.9
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=40, met=1.3, clo=0.3, body_position="sitting"
        )["m_rsw"]
        == 126.8
    )
    assert (
        use_fans_heatwaves(
            tdb=39, tr=39, v=0.2, rh=40, met=1.3, clo=0.5, body_position="sitting"
        )["e_rsw"]
        == 84.9
    )



import numpy as np

from pythermalcomfort.utilities import (
    units_converter,
)


def clo_tout(tout, units="SI"):
    """Representative clothing insulation Icl as a function of outdoor air
    temperature at 06:00 a.m [4]_.

    Parameters
    ----------
    tout : float or array-like
        outdoor air temperature at 06:00 a.m., default in [°C] in [°F] if `units` = 'IP'
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    clo : float or array-like
         Representative clothing insulation Icl, [clo]

    Notes
    -----
    The ASHRAE 55 2020 states that it is acceptable to determine the clothing
    insulation Icl using this equation in mechanically conditioned buildings [1]_.

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import clo_tout
        >>> clo_tout(tout=27)
        0.46
        >>> clo_tout(tout=[27, 25])
        array([0.46, 0.47])
    """

    tout = np.array(tout)

    if units.lower() == "ip":
        tout = units_converter(tmp=tout)[0]

    clo = np.where(tout < 26, np.power(10, -0.1635 - 0.0066 * tout), 0.46)
    clo = np.where(tout < 5, 0.818 - 0.0364 * tout, clo)
    clo = np.where(tout < -5, 1, clo)

    return np.around(clo, 2)



def net(tdb, rh, v, **kwargs):
    """Calculates the Normal Effective Temperature (NET). Missenard (1933)
    devised a formula for calculating effective temperature. The index
    establishes a link between the same condition of the organism's
    thermoregulatory capability (warm and cold perception) and the surrounding
    environment's temperature and humidity. The index is calculated as a
    function of three meteorological factors: air temperature, relative
    humidity of air, and wind speed. This index allows to calculate the
    effective temperature felt by a person. Missenard original equation was
    then used to calculate the Normal Effective Temperature (NET), by
    considering normal atmospheric pressure and a normal human body temperature
    (37°C). The NET is still in use in Germany, where medical check-ups for
    subjects working in the heat are decided on by prevailing levels of ET,
    depending on metabolic rates. The NET is also constantly monitored by the
    Hong Kong Observatory [16]_. In central Europe the following thresholds are
    in use: <1°C = very cold; 1–9 = cold; 9–17 = cool; 17–21 = fresh; 21–23 = comfortable;
    23–27 = warm; >27°C = hot [16]_.

    Parameters
    ----------
    tdb : float,
        dry bulb air temperature, [°C]
    rh : float
        relative humidity, [%]
    v : float
        wind speed [m/s] at 1.2 m above the ground

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output value, if False it does not round it

    Returns
    -------
    net : float
        Normal Effective Temperature, [°C]

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import net
        >>> net(tdb=37, rh=100, v=0.1)
        37
    """
    default_kwargs = {
        "round": True,
    }
    kwargs = {**default_kwargs, **kwargs}

    frac = 1.0 / (1.76 + 1.4 * v**0.75)

    et = 37 - (37 - tdb) / (0.68 - 0.0014 * rh + frac) - 0.29 * tdb * (1 - 0.01 * rh)

    if kwargs["round"]:
        return round(et, 1)
    else:
        return et



import numpy as np

from pythermalcomfort.models import e_pmv


def test_e_pmv():
    np.testing.assert_equal(
        e_pmv([24, 30], 30, vr=0.22, rh=50, met=1.4, clo=0.5, e_coefficient=0.6),
        [0.29, 0.91],
    )



import numpy as np

from pythermalcomfort.models.pmv_ppd import _pmv_ppd_optimized


def athb(tdb, tr, vr, rh, met, t_running_mean):
    """Return the PMV value calculated with the Adaptive Thermal Heat Balance
    Framework [27]_. The adaptive thermal heat balance (ATHB) framework
    introduced a method to account for the three adaptive principals, namely
    physiological, behavioral, and psychological adaptation, individually
    within existing heat balance models. The objective is a predictive model of
    thermal sensation applicable during the design stage or in international
    standards without knowing characteristics of future occupants.

    Parameters
    ----------
    tdb : float or array-like
        dry bulb air temperature, in [°C]
    tr : float or array-like
        mean radiant temperature, in [°C]
    vr : float or array-like
        relative air speed, in [m/s]

        Note: vr is the relative air speed caused by body movement and not the air
        speed measured by the air speed sensor. The relative air speed is the sum of the
        average air speed measured by the sensor plus the activity-generated air speed
        (Vag). Where Vag is the activity-generated air speed caused by motion of
        individual body parts. vr can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.v_relative`.
    rh : float or array-like
        relative humidity, [%]
    met : float or array-like
        metabolic rate, [met]
    t_running_mean: float or array-like
        running mean temperature, in [°C]

        The running mean temperature can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.running_mean_outdoor_temperature`.

    Returns
    -------
    athb_pmv : float or array-like
        Predicted Mean Vote calculated with the Adaptive Thermal Heat Balance framework

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import athb
        >>> print(athb( tdb=[25, 27], tr=25, vr=0.1, rh=50, met=1.1, t_running_mean=20))
        [0.2, 0.209]
    """
    tdb = np.array(tdb)
    tr = np.array(tr)
    vr = np.array(vr)
    met = np.array(met)
    rh = np.array(rh)
    t_running_mean = np.array(t_running_mean)

    met_adapted = met - (0.234 * t_running_mean) / 58.2

    # adapted clothing insulation level through behavioural adaptation
    clo_adapted = np.power(
        10,
        (
            -0.17168
            - 0.000485 * t_running_mean
            + 0.08176 * met_adapted
            - 0.00527 * t_running_mean * met_adapted
        ),
    )

    pmv_res = _pmv_ppd_optimized(tdb, tr, vr, rh, met_adapted, clo_adapted, 0)
    ts = 0.303 * np.exp(-0.036 * met_adapted * 58.15) + 0.028
    l_adapted = pmv_res / ts

    # predicted thermal sensation vote
    return np.around(
        1.484
        + 0.0276 * l_adapted
        - 0.9602 * met_adapted
        - 0.0342 * t_running_mean
        + 0.0002264 * l_adapted * t_running_mean
        + 0.018696 * met_adapted * t_running_mean
        - 0.0002909 * l_adapted * met_adapted * t_running_mean,
        3,
    )



import numpy as np

from pythermalcomfort.models import adaptive_ashrae


def test_adaptive_ashrae():
    data_test_adaptive_ashrae = (
        [  # I have commented the lines of code that don't pass the test
            {
                "tdb": 19.6,
                "tr": 19.6,
                "t_running_mean": 17,
                "v": 0.1,
                "return": {"acceptability_80": True},
            },
            {
                "tdb": 19.6,
                "tr": 19.6,
                "t_running_mean": 17,
                "v": 0.1,
                "return": {"acceptability_90": False},
            },
            {
                "tdb": 19.6,
                "tr": 19.6,
                "t_running_mean": 25,
                "v": 0.1,
                "return": {"acceptability_80": False},
            },
            {
                "tdb": 19.6,
                "tr": 19.6,
                "t_running_mean": 25,
                "v": 0.1,
                "return": {"acceptability_80": False},
            },
            {
                "tdb": 26,
                "tr": 26,
                "t_running_mean": 16,
                "v": 0.1,
                "return": {"acceptability_80": True},
            },
            {
                "tdb": 26,
                "tr": 26,
                "t_running_mean": 16,
                "v": 0.1,
                "return": {"acceptability_90": False},
            },
            {
                "tdb": 30,
                "tr": 26,
                "t_running_mean": 16,
                "v": 0.1,
                "return": {"acceptability_80": False},
            },
            {
                "tdb": 25,
                "tr": 25,
                "t_running_mean": 23,
                "v": 0.1,
                "return": {"acceptability_80": True},
            },
            {
                "tdb": 25,
                "tr": 25,
                "t_running_mean": 23,
                "v": 0.1,
                "return": {"acceptability_90": True},
            },
        ]
    )
    for row in data_test_adaptive_ashrae:
        assert (
            adaptive_ashrae(row["tdb"], row["tr"], row["t_running_mean"], row["v"])[
                list(row["return"].keys())[0]
            ]
        ) == row["return"][list(row["return"].keys())[0]]

    assert (adaptive_ashrae(77, 77, 68, 0.3, units="ip")["tmp_cmf"]) == 75.2

    # test limit_inputs and array input
    np.testing.assert_equal(
        adaptive_ashrae(tdb=25, tr=25, t_running_mean=[9, 10], v=0.1).__dict__,
        {
            "tmp_cmf": [np.nan, 20.9],
            "tmp_cmf_80_low": [np.nan, 17.4],
            "tmp_cmf_80_up": [np.nan, 24.4],
            "tmp_cmf_90_low": [np.nan, 18.4],
            "tmp_cmf_90_up": [np.nan, 23.4],
            "acceptability_80": [False, False],
            "acceptability_90": [False, False],
        },
    )
    np.testing.assert_equal(
        adaptive_ashrae(
            tdb=[77, 74], tr=77, t_running_mean=[48, 68], v=0.3, units="ip"
        ).__dict__,
        {
            "tmp_cmf": [np.nan, 75.2],
            "tmp_cmf_80_low": [np.nan, 68.9],
            "tmp_cmf_80_up": [np.nan, 81.5],
            "tmp_cmf_90_low": [np.nan, 70.7],
            "tmp_cmf_90_up": [np.nan, 79.7],
            "acceptability_80": [False, True],
            "acceptability_90": [False, True],
        },
    )



from pythermalcomfort.models import wc


def test_wc():
    assert wc(tdb=0, v=0.1) == {"wci": 518.6}
    assert wc(tdb=0, v=1.5) == {"wci": 813.5}
    assert wc(tdb=-5, v=5.5) == {"wci": 1255.2}
    assert wc(tdb=-10, v=11) == {"wci": 1631.1}
    assert wc(tdb=-5, v=11) == {"wci": 1441.4}



import math

from pythermalcomfort.utilities import (
    units_converter,
    check_standard_compliance,
)
from pythermalcomfort.models import pmv


def vertical_tmp_grad_ppd(tdb, tr, vr, rh, met, clo, vertical_tmp_grad, units="SI"):
    """Calculates the percentage of thermally dissatisfied people with a
    vertical temperature gradient between feet and head [1]_. This equation is
    only applicable for vr < 0.2 m/s (40 fps).

    Parameters
    ----------
    tdb : float
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'

        Note: The air temperature is the average value over two heights: 0.6 m (24 in.)
        and 1.1 m (43 in.) for seated occupants
        and 1.1 m (43 in.) and 1.7 m (67 in.) for standing occupants.
    tr : float
        mean radiant temperature, default in [°C] in [°F] if `units` = 'IP'
    vr : float
        relative air speed, default in [m/s] in [fps] if `units` = 'IP'

        Note: vr is the relative air speed caused by body movement and not the air
        speed measured by the air speed sensor. The relative air speed is the sum of the
        average air speed measured by the sensor plus the activity-generated air speed
        (Vag). Where Vag is the activity-generated air speed caused by motion of
        individual body parts. vr can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.v_relative`.
    rh : float
        relative humidity, [%]
    met : float
        metabolic rate, [met]
    clo : float
        clothing insulation, [clo]

        Note: The activity as well as the air speed modify the insulation characteristics
        of the clothing and the adjacent air layer. Consequently the ISO 7730 states that
        the clothing insulation shall be corrected [2]_. The ASHRAE 55 Standard corrects
        for the effect of the body movement for met equal or higher than 1.2 met using
        the equation clo = Icl × (0.6 + 0.4/met) The dynamic clothing insulation, clo,
        can be calculated using the function
        :py:meth:`pythermalcomfort.utilities.clo_dynamic`.
    vertical_tmp_grad : float
        vertical temperature gradient between the feet and the head, default in [°C/m]
        in [°F/ft] if `units` = 'IP'
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    PPD_vg: float
        Predicted Percentage of Dissatisfied occupants with vertical temperature
        gradient, [%]
    Acceptability: bol
        The ASHRAE 55 2020 standard defines that the value of air speed at the ankle
        level is acceptable if PPD_ad is lower or equal than 5 %

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import vertical_tmp_grad_ppd
        >>> results = vertical_tmp_grad_ppd(25, 25, 0.1, 50, 1.2, 0.5, 7)
        >>> print(results)
        {'PPD_vg': 12.6, 'Acceptability': False}
    """
    if units.lower() == "ip":
        tdb, tr, vr = units_converter(tdb=tdb, tr=tr, v=vr)
        vertical_tmp_grad = vertical_tmp_grad / 1.8 * 3.28

    check_standard_compliance(
        standard="ashrae", tdb=tdb, tr=tr, v_limited=vr, rh=rh, met=met, clo=clo
    )

    tsv = pmv(tdb, tr, vr, rh, met, clo, standard="ashrae")
    numerator = math.exp(0.13 * (tsv - 1.91) ** 2 + 0.15 * vertical_tmp_grad - 1.6)
    ppd_val = round((numerator / (1 + numerator) - 0.345) * 100, 1)
    acceptability = ppd_val <= 5
    return {"PPD_vg": ppd_val, "Acceptability": acceptability}



import numpy as np

from pythermalcomfort.utilities import (
    mapping,
)


def discomfort_index(tdb, rh):
    """Calculates the Discomfort Index (DI). The index is essentially an
    effective temperature based on air temperature and humidity. The discomfort
    index is usuallly divided in 6 dicomfort categories and it only applies to
    warm environments: [24]_

    * class 1 - DI < 21 °C - No discomfort
    * class 2 - 21 <= DI < 24 °C - Less than 50% feels discomfort
    * class 3 - 24 <= DI < 27 °C - More than 50% feels discomfort
    * class 4 - 27 <= DI < 29 °C - Most of the population feels discomfort
    * class 5 - 29 <= DI < 32 °C - Everyone feels severe stress
    * class 6 - DI >= 32 °C - State of medical emergency

    Parameters
    ----------
    tdb : float or array-like
        dry bulb air temperature, [°C]
    rh : float or array-like
        relative humidity, [%]

    Returns
    -------
    di : float or array-like
        Discomfort Index, [°C]
    discomfort_condition : str or array-like
        Classification of the thermal comfort conditions according to the discomfort index

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import discomfort_index
        >>> discomfort_index(tdb=25, rh=50)
        {'di': 22.1, 'discomfort_condition': 'Less than 50% feels discomfort'}
    """

    tdb = np.array(tdb)
    rh = np.array(rh)

    di = tdb - 0.55 * (1 - 0.01 * rh) * (tdb - 14.5)

    di_categories = {
        21: "No discomfort",
        24: "Less than 50% feels discomfort",
        27: "More than 50% feels discomfort",
        29: "Most of the population feels discomfort",
        32: "Everyone feels severe stress",
        99: "State of medical emergency",
    }

    return {
        "di": np.around(di, 1),
        "discomfort_condition": mapping(di, di_categories, right=False),
    }



import pandas as pd
import matplotlib.pyplot as plt
import os
from pythermalcomfort import atcs

# Make "atcs_output_example" directory in the current directory
directory_name = "atcs_output_example"
current_directory = os.getcwd()
atcs_example_directory = os.path.join(current_directory, directory_name)
if not os.path.exists(atcs_example_directory):
    os.makedirs(atcs_example_directory)

# -------------------------------------------
# EXAMPLE (simple simulation)
# The following code simulates human thermal sensation in transient environment.
# The simulated people is exposed to a warm environment and goes to cool environment.
# This model can simulate the sudden changes in thermal sensations we usually experience
# when the environmental step change happens.
# -------------------------------------------

# Build a model and set a body built
# Create an instance of the class with optional body parameters such as body height, weight, age, sex, etc.
model = atcs.ATCS(height=1.7, weight=60, age=30)

# Set the first phase
model.to = 28  # Operative temperature [°C]
model.rh = 40  # Relative humidity [%]
model.v = 0.2  # Air velocity [m/s]
model.par = 1.2  # Physical activity ratio [-]
model.simulate(60)  # Exposure time = 60 [min]

# Set the next condition (You only need to change the parameters that you want to change)
model.to = 20  # Change only operative temperature
model.simulate(60)  # Additional exposure time = 60 [min]

# Show the results
df = pd.DataFrame(model.dict_results())
df["sensation_overall"].plot()
plt.ylabel("Overall thermal sensation [-]")
plt.xlabel("Time [min]")
plt.savefig(os.path.join(atcs_example_directory, "atcs_example_local_sensations.png"))
plt.show()

# Exporting the results as csv
model.to_csv(os.path.join(atcs_example_directory, "atcs_example1 (default output).csv"))



def humidex(tdb, rh, **kwargs):
    """Calculates the humidex (short for "humidity index"). It has been
    developed by the Canadian Meteorological service. It was introduced in 1965
    and then it was revised by Masterson and Richardson (1979) [14]_. It aims
    to describe how hot, humid weather is felt by the average person. The
    Humidex differs from the heat index in being related to the dew point
    rather than relative humidity [15]_.

    Parameters
    ----------
    tdb : float
        dry bulb air temperature, [°C]
    rh : float
        relative humidity, [%]

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output value, if False it does not round it

    Returns
    -------
    humidex: float
        Heat Index, [°C]
    discomfort: str
        Degree of Comfort or Discomfort as defined in Havenith and Fiala (2016) [15]_

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import humidex
        >>> humidex(tdb=25, rh=50)
        {"humidex": 28.2, "discomfort": "Little or no discomfort"}
    """
    default_kwargs = {
        "round": True,
    }
    kwargs = {**default_kwargs, **kwargs}

    hi = tdb + 5 / 9 * ((6.112 * 10 ** (7.5 * tdb / (237.7 + tdb)) * rh / 100) - 10)

    if kwargs["round"]:
        hi = round(hi, 1)

    stress_category = "Heat stroke probable"
    if hi <= 30:
        stress_category = "Little or no discomfort"
    elif hi <= 35:
        stress_category = "Noticeable discomfort"
    elif hi <= 40:
        stress_category = "Evident discomfort"
    elif hi <= 45:
        stress_category = "Intense discomfort; avoid exertion"
    elif hi <= 54:
        stress_category = "Dangerous discomfort"

    return {"humidex": hi, "discomfort": stress_category}



def heat_index(tdb, rh, **kwargs):
    """Calculates the Heat Index (HI). It combines air temperature and relative
    humidity to determine an apparent temperature. The HI equation [12]_ is
    derived by multiple regression analysis in temperature and relative
    humidity from the first version of Steadman’s (1979) apparent temperature
    (AT) [13]_.

    Parameters
    ----------
    tdb : float
        dry bulb air temperature, default in [°C] in [°F] if `units` = 'IP'
    rh : float
        relative humidity, [%]

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output value, if False it does not round it
    units : {'SI', 'IP'}
        select the SI (International System of Units) or the IP (Imperial Units) system.

    Returns
    -------
    hi : float
        Heat Index, default in [°C] in [°F] if `units` = 'IP'

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import heat_index
        >>> heat_index(tdb=25, rh=50)
        25.9
    """
    default_kwargs = {
        "round": True,
        "units": "SI",
    }
    kwargs = {**default_kwargs, **kwargs}

    if kwargs["units"] == "SI":
        # from doi: 10.1007/s00484-011-0453-2
        hi = -8.784695 + 1.61139411 * tdb + 2.338549 * rh - 0.14611605 * tdb * rh
        hi += -1.2308094 * 10**-2 * tdb**2 - 1.6424828 * 10**-2 * rh**2
        hi += 2.211732 * 10**-3 * tdb**2 * rh + 7.2546 * 10**-4 * tdb * rh**2
        hi += -3.582 * 10**-6 * tdb**2 * rh**2

    else:
        # from doi: 10.1007/s00484-021-02105-0
        hi = -42.379 + 2.04901523 * tdb + 10.14333127 * rh
        hi += -0.22475541 * tdb * rh - 6.83783 * 10**-3 * tdb**2
        hi += -5.481717 * 10**-2 * rh**2
        hi += 1.22874 * 10**-3 * tdb**2 * rh + 8.5282 * 10**-4 * tdb * rh**2
        hi += -1.99 * 10**-6 * tdb**2 * rh**2

    if kwargs["round"]:
        return round(hi, 1)
    else:
        return hi



def wc(tdb, v, **kwargs):
    """Calculates the Wind Chill Index (WCI) in accordance with the ASHRAE 2017 Handbook Fundamentals - Chapter 9 [18]_.

    The wind chill index (WCI) is an empirical index based on cooling measurements
    taken on a cylindrical flask partially filled with water in Antarctica
    (Siple and Passel 1945). For a surface temperature of 33°C, the index describes
    the rate of heat loss from the cylinder via radiation and convection as a function
    of ambient temperature and wind velocity.

    This formulation has been met with some valid criticism. WCI is unlikely to be an
    accurate measure of heat loss from exposed flesh, which differs from plastic in terms
    of curvature, roughness, and radiation exchange qualities, and is always below 33°C
    in a cold environment. Furthermore, the equation's values peak at 90 km/h and then
    decline as velocity increases. Nonetheless, this score reliably represents the
    combined effects of temperature and wind on subjective discomfort for velocities
    below 80 km/h [18]_.

    Parameters
    ----------
    tdb : float
        dry bulb air temperature,[°C]
    v : float
        wind speed 10m above ground level, [m/s]

    Other Parameters
    ----------------
    round: boolean, default True
        if True rounds output value, if False it does not round it

    Returns
    -------
    wci: float
        wind chill index, [W/m2)]

    Examples
    --------
    .. code-block:: python

        >>> from pythermalcomfort.models import wc
        >>> wc(tdb=-5, v=5.5)
        {"wci": 1255.2}
    """
    default_kwargs = {
        "round": True,
    }
    kwargs = {**default_kwargs, **kwargs}

    wci = (10.45 + 10 * v**0.5 - v) * (33 - tdb)

    # the factor 1.163 is used to convert to W/m2
    wci = wci * 1.163

    if kwargs["round"]:
        wci = round(wci, 1)

    return {"wci": wci}
